"""protocol.py

Created by Thomas Mangin on 2009-08-25.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from __future__ import annotations

import os
import traceback
from typing import TYPE_CHECKING, Any, cast
from collections.abc import AsyncGenerator

if TYPE_CHECKING:
    from exabgp.bgp.neighbor import Neighbor
    from exabgp.reactor.network.incoming import Incoming
    from exabgp.reactor.peer import Peer

# ================================================================ Registration
#

from exabgp.bgp.message import _NOP, EOR, KeepAlive, Message, Notification, Notify, Open, Operational, Update
from exabgp.bgp.message.direction import Direction
from exabgp.bgp.message.open import RouterID, Version
from exabgp.bgp.message.open.capability import Capabilities, Negotiated
from exabgp.bgp.message.refresh import RouteRefresh
from exabgp.bgp.message.update import UpdateCollection
from exabgp.bgp.message.update.attribute import Attribute, AttributeCollection
from exabgp.logger import lazymsg, log

# from exabgp.reactor.network.error import NotifyError
from exabgp.protocol.family import AFI, SAFI
from exabgp.protocol.ip import IP
from exabgp.reactor.network.outgoing import Outgoing

# This is the number of chuncked message we are willing to buffer, not the number of routes
MAX_BACKLOG = 15000

_UPDATE = UpdateCollection([], [], AttributeCollection())
_OPERATIONAL = Operational(0x00)


class Protocol:
    decode: bool = True

    def __init__(self, peer: 'Peer') -> None:
        self.peer: 'Peer' = peer
        self.neighbor: 'Neighbor' = peer.neighbor
        self.negotiated: Negotiated = Negotiated.make_negotiated(self.neighbor, Direction.IN)
        self.connection: 'Incoming' | Outgoing | None = None

        if self.neighbor.session.connect:
            self.port: int = self.neighbor.session.connect
        elif os.environ.get('exabgp.tcp.port', '').isdigit():
            self.port = int(os.environ['exabgp.tcp.port'])
        elif os.environ.get('exabgp_tcp_port', '').isdigit():
            self.port = int(os.environ['exabgp_tcp_port'])
        else:
            self.port = 179

        from exabgp.environment import getenv

        self.log_routes: bool = peer.neighbor.adj_rib_in or getenv().log.routes

    def fd(self) -> int:
        if self.connection is None:
            return -1
        return self.connection.fd()

    def _session(self) -> str:
        """Return session identifier for logging. Requires connection to be established."""
        assert self.connection is not None
        return self.connection.session()

    @property
    def _api(self) -> dict[str, Any]:
        """Return neighbor API config."""
        return self.neighbor.api

    # Note: We use self.peer.neighbor for consistency - both reference the same object
    # but self.peer.neighbor is used throughout to maintain clear ownership semantics.

    def me(self, message: str) -> str:
        return f'{self.peer.neighbor.session.peer_address}/{self.peer.neighbor.session.peer_as} {message}'

    def accept(self, incoming: 'Incoming') -> Protocol:
        self.connection = incoming

        if self._api['neighbor-changes']:
            self.peer.reactor.processes.connected(self.peer.neighbor)

        # very important - as we use this function on __init__
        return self

    async def connect(self) -> bool:
        """Establish connection using asyncio.

        Returns:
            True if connection successful, False otherwise
        """
        # allows to test the protocol code using modified StringIO with a extra 'pending' function
        if self.connection:
            return True

        assert self.neighbor.session.peer_address is not None
        local = (
            self.neighbor.session.md5_ip.top()
            if not self.neighbor.session.auto_discovery and self.neighbor.session.md5_ip
            else ''
        )
        peer = self.neighbor.session.peer_address.top()
        afi = self.neighbor.session.peer_address.afi
        md5 = self.neighbor.session.md5_password
        md5_base64 = self.neighbor.session.md5_base64
        ttl_out = self.neighbor.session.outgoing_ttl
        itf = self.neighbor.session.source_interface
        # TCP-AO (RFC 5925)
        tcp_ao_keyid = self.neighbor.session.tcp_ao_keyid
        tcp_ao_algorithm = self.neighbor.session.tcp_ao_algorithm
        tcp_ao_password = self.neighbor.session.tcp_ao_password
        tcp_ao_base64 = self.neighbor.session.tcp_ao_base64
        self.connection = Outgoing(
            afi,
            peer,
            local,
            self.port,
            md5,
            md5_base64,
            ttl_out,
            itf,
            tcp_ao_keyid,
            tcp_ao_algorithm,
            tcp_ao_password,
            tcp_ao_base64,
        )

        # Use async establish instead of generator
        connected = await self.connection.establish_async()

        if not connected:
            return False

        if self._api['neighbor-changes']:
            self.peer.reactor.processes.connected(self.peer.neighbor)

        if not local:
            self.neighbor.session.local_address = IP.from_string(self.connection.local)
            if self.neighbor.session.router_id is None and self.neighbor.session.local_address.afi == AFI.ipv4:
                self.neighbor.session.router_id = RouterID(self.neighbor.session.local_address.top())

        return True

    def close(self, reason: str = 'protocol closed, reason unspecified') -> None:
        if self.connection:
            log.debug(lazymsg('protocol.close reason={r}', r=reason), self._session())
            self.peer.stats['down'] += 1

            self.connection.close()
            self.connection = None

    def _to_api(self, direction: str, message: Any, raw: bytes) -> None:
        packets: bool = self._api['{}-packets'.format(direction)]
        parsed: bool = self._api['{}-parsed'.format(direction)]
        consolidate: bool = self._api['{}-consolidate'.format(direction)]
        neg: Negotiated = self.negotiated

        if consolidate:
            if packets:
                self.peer.reactor.processes.message(
                    message.ID, self.peer, direction, message, raw[:19], raw[19:], negotiated=neg
                )
            else:
                self.peer.reactor.processes.message(message.ID, self.peer, direction, message, b'', b'', negotiated=neg)
        else:
            if packets:
                self.peer.reactor.processes.packets(
                    self.peer.neighbor, direction, int(message.ID), raw[:19], raw[19:], neg
                )
            if parsed:
                self.peer.reactor.processes.message(message.ID, self.peer, direction, message, b'', b'', negotiated=neg)

    async def write(self, message: Any, negotiated: Negotiated) -> None:
        """Send BGP message using async I/O."""
        assert self.connection is not None
        raw: bytes = message.pack_message(negotiated)

        code: str = 'send-{}'.format(Message.CODE.short(message.ID))
        self.peer.stats[code] += 1
        if self._api.get(code, False):
            self._to_api('send', message, raw)

        await self.connection.writer_async(raw)

    async def send(self, raw: bytes) -> None:
        """Send raw BGP message using async I/O."""
        assert self.connection is not None
        code: str = 'send-{}'.format(Message.CODE.short(raw[18]))
        self.peer.stats[code] += 1
        if self._api.get(code, False):
            # Parse the raw bytes to get an Update for API
            update = Update(raw[19:])
            update.parse(self.negotiated)
            self._to_api('send', update, raw)

        await self.connection.writer_async(raw)

    # Read from network .......................................................

    async def read_message(self) -> Message:
        """Read BGP message using async I/O."""
        assert self.connection is not None
        packets = self._api['receive-packets']
        consolidate = self._api['receive-consolidate']
        parsed = self._api['receive-parsed']

        # Read message using async I/O
        length, msg_id, header, body, notify = await self.connection.reader_async()

        # internal issue
        if notify:
            code = 'receive-{}'.format(Message.CODE.NOTIFICATION.SHORT)
            # Convert NotifyError to Notify for API and exception
            notify_msg = Notify(notify.code, notify.subcode, str(notify))
            if self._api.get(code, False):
                if consolidate:
                    self.peer.reactor.processes.notification(
                        self.peer.neighbor, 'receive', notify_msg, bytes(header), bytes(body), self.negotiated
                    )
                elif parsed:
                    self.peer.reactor.processes.notification(
                        self.peer.neighbor, 'receive', notify_msg, b'', b'', self.negotiated
                    )
                elif packets:
                    self.peer.reactor.processes.packets(
                        self.peer.neighbor, 'receive', msg_id, bytes(header), bytes(body), self.negotiated
                    )
            raise notify_msg

        if msg_id not in Message.CODE.MESSAGES:
            raise Notify(1, 0, 'can not decode update message of type "%d"' % msg_id)

        if not length:
            return _NOP

        current_msg_id = msg_id
        log.debug(
            lazymsg('message.received type={t}', t=Message.CODE.name(current_msg_id)),
            self._session(),
        )

        code = 'receive-{}'.format(Message.CODE.short(msg_id))
        self.peer.stats[code] += 1
        for_api = self._api.get(code, False)

        if for_api and packets and not consolidate:
            self.peer.reactor.processes.packets(
                self.peer.neighbor, 'receive', msg_id, bytes(header), bytes(body), self.negotiated
            )

        if msg_id == Message.CODE.UPDATE:
            if not self.neighbor.adj_rib_in and not (for_api or self.log_routes) and not (parsed or consolidate):
                return _UPDATE

        try:
            message = Message.unpack(msg_id, body, self.negotiated)
        except (KeyboardInterrupt, SystemExit, Notify):
            raise
        except Exception as exc:
            current_msg_id = msg_id
            log.debug(lazymsg('message.decode.failed type={t}', t=current_msg_id), self._session())
            current_exc = exc
            log.debug(lazymsg('message.decode.error error={e}', e=str(current_exc)), self._session())
            log.debug(lazymsg('message.decode.traceback trace={t}', t=traceback.format_exc()), self._session())
            raise Notify(1, 0, 'can not decode update message of type "%d"' % msg_id) from None
            # raise Notify(5,0,'unknown message received')

        if message.TYPE == Update.TYPE:
            # Both Update and EOR have TYPE == Update.TYPE
            # Update: use .data to get parsed collection
            # EOR: has .attributes and .nlris directly
            if isinstance(message, Update):
                # Note: TREAT_AS_WITHDRAW handling is done at the Update level
                # The UpdateCollection (message.data) already tracks which NLRIs
                # are withdraws via the announces vs withdraws lists
                pass

        if for_api:
            if consolidate:
                self.peer.reactor.processes.message(
                    msg_id, self.peer, 'receive', message, bytes(header), bytes(body), self.negotiated
                )
            elif parsed:
                self.peer.reactor.processes.message(msg_id, self.peer, 'receive', message, b'', b'', self.negotiated)

        if message.TYPE == Notification.TYPE:
            raise cast(Notification, message)

        if isinstance(message, Update) and Attribute.CODE.INTERNAL_DISCARD in message.data.attributes:
            return _NOP
        else:
            return message

    def validate_open(self) -> None:
        error: tuple[int, int, str] | None = self.negotiated.validate(self.neighbor)
        if error is not None:
            raise Notify(*error)

        if self._api['negotiated']:
            self.peer.reactor.processes.negotiated(self.peer.neighbor, self.negotiated)

        if self.negotiated.mismatch and self.connection is not None:
            log.warning(
                lazymsg('negotiation.family.mismatch count={c}', c=len(self.negotiated.mismatch)),
                self._session(),
            )
            for reason, (afi, safi) in self.negotiated.mismatch:
                current_afi, current_reason, current_safi = afi, reason, safi
                log.warning(
                    lazymsg(
                        'negotiation.family.unconfigured reason={r} afi={a} safi={s}',
                        r=current_reason,
                        a=current_afi,
                        s=current_safi,
                    ),
                    self._session(),
                )

    async def read_open(self, ip: str) -> Open:
        """Read OPEN message using async I/O."""
        while True:
            received_open = await self.read_message()
            if not received_open.SCHEDULING:  # Real message (not NOP)
                break

        if received_open.TYPE != Open.TYPE:
            raise Notify(
                5,
                1,
                'The first packet received is not an open message ({})'.format(received_open),
            )

        log.debug(lazymsg('open.received message={m}', m=received_open), self._session())
        return cast(Open, received_open)

    async def read_keepalive(self) -> KeepAlive:
        """Read KEEPALIVE message using async I/O."""
        while True:
            message = await self.read_message()
            if not message.SCHEDULING:  # Real message (not NOP)
                break

        if message.TYPE != KeepAlive.TYPE:
            raise Notify(5, 2)

        return cast(KeepAlive, message)

    #
    # Sending message to peer
    #

    async def new_open(self) -> Open:
        """Create and send OPEN message using async I/O."""
        assert self.connection is not None
        assert self.neighbor.session.router_id is not None
        if self.neighbor.session.local_as:
            local_as = self.neighbor.session.local_as
        elif self.negotiated.received_open:
            local_as = self.negotiated.received_open.asn
        else:
            raise RuntimeError('no ASN available for the OPEN message')

        sent_open = Open.make_open(
            Version(4),
            local_as,
            self.neighbor.hold_time,
            self.neighbor.session.router_id,
            Capabilities().new(self.neighbor, self.peer._restarted),
        )

        # we do not buffer open message in purpose
        await self.write(sent_open, self.negotiated)

        log.debug(lazymsg('open.sent message={m}', m=sent_open), self._session())
        return sent_open

    async def new_keepalive(self, comment: str = '') -> KeepAlive:
        """Create and send KEEPALIVE message using async I/O."""
        assert self.connection is not None
        keepalive: KeepAlive = KeepAlive()

        await self.write(keepalive, self.negotiated)

        log.debug(
            lazymsg('keepalive.sent comment={c}', c=comment if comment else 'none'),
            self._session(),
        )

        return keepalive

    async def new_notification(self, notification: Notify) -> Notify:
        """Send BGP NOTIFICATION message."""
        assert self.connection is not None
        await self.write(notification, self.negotiated)
        log.debug(
            lazymsg(
                'notification.sent code={c} subcode={sc} data={d}',
                c=notification.code,
                sc=notification.subcode,
                d=notification.data.decode('utf-8'),
            ),
            self._session(),
        )
        return notification

    async def new_update_generator(self, include_withdraw: bool) -> AsyncGenerator[None, None]:
        """Async generator for sending UPDATE messages - yields control between messages.

        This yields after each message to allow the event loop to process other tasks.
        """
        assert self.connection is not None
        log.debug(lazymsg('update.generator.started'), self._session())
        updates = self.neighbor.rib.outgoing.updates(
            self.neighbor.group_updates,
            paths_limit=self.negotiated.paths_limit or None,
        )
        number: int = 0
        for update in updates:
            for message in update.messages(self.negotiated, include_withdraw):
                number += 1
                current_msg = message
                log.debug(
                    lazymsg('update.message.sending num={num} msg={msg}', num=number, msg=repr(current_msg)),
                    self._session(),
                )
                await self.send(message)
                yield
        if number:
            final_number = number
            log.debug(lazymsg('update.sent count={n}', n=final_number), self._session())
        log.debug(lazymsg('update.generator.completed count={count}', count=number), self._session())

    async def new_update(self, include_withdraw: bool) -> UpdateCollection:
        """Send BGP UPDATE messages (runs to completion)."""
        assert self.connection is not None
        log.debug(lazymsg('update.started'), self._session())
        updates = self.neighbor.rib.outgoing.updates(
            self.neighbor.group_updates,
            paths_limit=self.negotiated.paths_limit or None,
        )
        number: int = 0
        for update in updates:
            current_update = update
            log.debug(
                lazymsg('update.processing update={upd}', upd=current_update),
                self._session(),
            )
            for message in update.messages(self.negotiated, include_withdraw):
                number += 1
                current_msg = message
                log.debug(
                    lazymsg('update.message.sending num={num} msg={msg}', num=number, msg=repr(current_msg)),
                    self._session(),
                )
                await self.send(message)
        if number:
            log.debug(lazymsg('update.sent count={n}', n=number), self._session())
        log.debug(lazymsg('update.completed count={count}', count=number), self._session())
        return _UPDATE

    async def new_eor(self, afi: AFI, safi: SAFI) -> EOR:
        """Send BGP End-of-RIB marker."""
        assert self.connection is not None
        eor: EOR = EOR(afi, safi)
        await self.write(eor, self.negotiated)
        log.debug(lazymsg('eor.sent afi={a} safi={s}', a=afi, s=safi), self._session())
        return eor

    async def new_eors(self, afi: AFI = AFI.undefined, safi: SAFI = SAFI.undefined) -> UpdateCollection:
        """Send End-of-RIB markers for all families."""
        if self.negotiated.families:
            families = self.negotiated.families if (afi, safi) == (AFI.undefined, SAFI.undefined) else [(afi, safi)]
            for eor_afi, eor_safi in families:
                await self.new_eor(eor_afi, eor_safi)
        else:
            # If not sending EOR, send keepalive
            await self.new_keepalive('EOR')
        return _UPDATE

    async def new_operational(self, operational: Operational, negotiated: Negotiated) -> Operational:
        """Send BGP OPERATIONAL message."""
        assert self.connection is not None
        await self.write(operational, negotiated)
        log.debug(lazymsg('operational.sent message={m}', m=str(operational)), self._session())
        return operational

    async def new_refresh(self, refresh: RouteRefresh) -> RouteRefresh:
        """Send BGP ROUTE-REFRESH message."""
        assert self.connection is not None
        await self.write(refresh, self.negotiated)
        log.debug(lazymsg('refresh.sent message={m}', m=str(refresh)), self._session())
        return refresh



"""shortcuts.py

Shared command shortcut/nickname expansion logic for ExaBGP CLI.

Created on 2025-11-20.
Copyright (c) 2009-2025 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from __future__ import annotations

from typing import Callable

from exabgp.protocol.ip import IPv4


# Type for shortcut matching function - can return bool or a truthy/falsy value
ShortcutMatcher = Callable[[int, list[str]], bool | list[str]]


def _announce_context(pos: int, pre: list[str]) -> bool:
    """Match 'a' → 'announce' when at start or after IP address."""
    return pos == 0 or (bool(pre) and (pre[-1].count('.') == IPv4.DOT_COUNT or ':' in pre[-1]))


def _attributes_context(pos: int, pre: list[str]) -> bool:
    """Match 'a' → 'attributes' after announce/withdraw."""
    return bool(pre) and (pre[-1] == 'announce' or pre[-1] == 'withdraw')


def _adj_rib_context(pos: int, pre: list[str]) -> bool:
    """Match 'a' → 'adj-rib' after clear/flush/show."""
    return bool(pre) and pre[-1] in ['clear', 'flush', 'show']


def _eor_context(pos: int, pre: list[str]) -> bool:
    """Match 'e' → 'eor' after announce."""
    return bool(pre) and pre[-1] == 'announce'


def _flow_context(pos: int, pre: list[str]) -> bool:
    """Match 'f' → 'flow' after announce/withdraw."""
    return bool(pre) and (pre[-1] == 'announce' or pre[-1] == 'withdraw')


def _flush_context(pos: int, pre: list[str]) -> bool:
    """Match 'f' → 'flush' at start or after IP address."""
    return pos == 0 or (bool(pre) and (pre[-1].count('.') == IPv4.DOT_COUNT or ':' in pre[-1]))


def _in_context(pos: int, pre: list[str]) -> bool:
    """Match 'i' → 'in' after adj-rib."""
    return bool(pre) and pre[-1] == 'adj-rib'


def _operation_context(pos: int, pre: list[str]) -> bool:
    """Match 'o' → 'operation' after announce."""
    return bool(pre) and pre[-1] == 'announce'


def _out_context(pos: int, pre: list[str]) -> bool:
    """Match 'o' → 'out' after adj-rib."""
    return bool(pre) and pre[-1] == 'adj-rib'


def _route_context(pos: int, pre: list[str]) -> bool:
    """Match 'r' → 'route' after announce/withdraw."""
    return bool(pre) and (pre[-1] == 'announce' or pre[-1] == 'withdraw')


def _teardown_context(pos: int, pre: list[str]) -> bool:
    """Match 't' → 'teardown' at start or after IP address."""
    return pos == 0 or (bool(pre) and (pre[-1].count('.') == IPv4.DOT_COUNT or ':' in pre[-1]))


def _vps_context(pos: int, pre: list[str]) -> bool:
    """Match 'v' → 'vps' after announce/withdraw."""
    return bool(pre) and (pre[-1] == 'announce' or pre[-1] == 'withdraw')


def _withdraw_context(pos: int, pre: list[str]) -> bool:
    """Match 'w' → 'withdraw' at start or after IP address."""
    return pos == 0 or (bool(pre) and (pre[-1].count('.') == IPv4.DOT_COUNT or ':' in pre[-1]))


def _watchdog_context(pos: int, pre: list[str]) -> bool:
    """Match 'w' → 'watchdog' after announce/withdraw."""
    return bool(pre) and (pre[-1] == 'announce' or pre[-1] == 'withdraw')


class CommandShortcuts:
    """Handles command shortcut expansion for CLI."""

    # Shortcut definitions: (nickname, full_name, match_condition)
    # Match condition is a function: (position: int, previous_tokens: list) -> bool
    SHORTCUTS: list[tuple[str, str, ShortcutMatcher]] = [
        # 'a' has multiple meanings based on context
        ('a', 'announce', _announce_context),
        ('a', 'attributes', _attributes_context),
        ('a', 'adj-rib', _adj_rib_context),
        ('a', 'ack', lambda pos, pre: len(pre) > 0 and pre[-1] == 'session'),  # session ack
        # Other single-letter shortcuts
        ('c', 'configuration', lambda pos, pre: True),
        ('d', 'daemon', lambda pos, pre: pos == 0),  # Top-level only
        ('e', 'eor', _eor_context),
        ('e', 'extensive', lambda pos, pre: 'show' in pre),
        ('f', 'flow', _flow_context),
        ('f', 'flush', _flush_context),
        ('h', 'help', lambda pos, pre: pos == 0),
        ('i', 'in', _in_context),
        ('id', 'router-id', lambda pos, pre: 'neighbor' in pre),  # CLI shortcut: id → router-id in neighbor context
        ('n', 'neighbor', lambda pos, pre: pos == 0),  # Only at start - removed 'show n' → 'show neighbor' support
        ('o', 'operation', _operation_context),
        ('o', 'out', _out_context),
        ('r', 'route', _route_context),
        ('r', 'refresh', lambda pos, pre: len(pre) >= 2 and pre[-2] == 'announce' and pre[-1] == 'route'),
        ('s', 'show', lambda pos, pre: pos == 0),
        ('s', 'summary', lambda pos, pre: pos != 0),
        ('t', 'teardown', _teardown_context),
        ('v', 'vps', _vps_context),
        ('w', 'withdraw', _withdraw_context),
        ('w', 'watchdog', _watchdog_context),
        # Multi-letter shortcuts
        ('ses', 'session', lambda pos, pre: pos == 0),  # Top-level only (avoid 's' conflict)
        ('sy', 'sync', lambda pos, pre: len(pre) > 0 and pre[-1] == 'session'),  # session sync
        # Common typos
        ('neighbour', 'neighbor', lambda pos, pre: True),
        ('neigbour', 'neighbor', lambda pos, pre: True),
        ('neigbor', 'neighbor', lambda pos, pre: True),
    ]

    @classmethod
    def expand_shortcuts(cls, command: str) -> str:
        """Expand shortcuts in a command string and apply CLI transformations.

        Args:
            command: The command string to process (e.g., "s n summary")

        Returns:
            Expanded and transformed command string (e.g., "show neighbor summary")

        Examples:
            >>> CommandShortcuts.expand_shortcuts('s n summary')
            'show neighbor summary'
            >>> CommandShortcuts.expand_shortcuts('a r 10.0.0.0/24 next-hop 192.168.1.1')
            'announce route 10.0.0.0/24 next-hop 192.168.1.1'
            >>> CommandShortcuts.expand_shortcuts('n 192.168.1.1 show summary')
            'show neighbor 192.168.1.1 summary'
        """
        tokens = command.split()
        if not tokens:
            return command

        expanded = cls.expand_token_list(tokens)
        expanded_str = ' '.join(expanded).strip()

        # Apply CLI-to-API transformations
        transformed = cls.transform_cli_to_api(expanded_str)
        return transformed

    @classmethod
    def expand_token_list(cls, tokens: list[str]) -> list[str]:
        """Expand shortcuts in a list of tokens.

        Args:
            tokens: List of command tokens

        Returns:
            List of expanded tokens

        Examples:
            >>> CommandShortcuts.expand_token_list(['s', 'n', 'summary'])
            ['show', 'neighbor', 'summary']
        """
        expanded: list[str] = []

        for pos, token in enumerate(tokens):
            # Try to match against shortcuts
            matched = False
            for nickname, full_name, match_condition in cls.SHORTCUTS:
                # Check if token matches:
                # 1. Exact nickname match (e.g., 'r' == 'r')
                # 2. Exact full_name match (e.g., 'route' == 'route')
                # 3. Prefix match for short tokens only (e.g., 'ro' matches 'route')
                #    This prevents 'route' from matching 'router-id'
                is_match = token == nickname or token == full_name or (len(token) <= 2 and full_name.startswith(token))
                if is_match and match_condition(pos, expanded):
                    expanded.append(full_name)
                    matched = True
                    break

            # If no shortcut matched, use the token as-is
            if not matched:
                expanded.append(token)

        return expanded

    @classmethod
    def get_expansion(cls, token: str, position: int, previous_tokens: list[str]) -> str:
        """Get the expansion for a single token in context.

        Args:
            token: The token to expand
            position: Position in command (0-indexed)
            previous_tokens: Previously expanded tokens

        Returns:
            Expanded token or original if no match

        Examples:
            >>> CommandShortcuts.get_expansion('s', 0, [])
            'show'
            >>> CommandShortcuts.get_expansion('a', 1, ['show'])
            'adj-rib'
            >>> CommandShortcuts.get_expansion('a', 0, [])
            'announce'
        """
        for nickname, full_name, match_condition in cls.SHORTCUTS:
            is_match = token == nickname or token == full_name or (len(token) <= 2 and full_name.startswith(token))
            if is_match and match_condition(position, previous_tokens):
                return full_name
        return token

    @classmethod
    def get_possible_expansions(cls, token: str, position: int, previous_tokens: list[str]) -> list[str]:
        """Get all possible expansions for a token in context.

        Useful for showing ambiguous shortcuts.

        Args:
            token: The token to expand
            position: Position in command (0-indexed)
            previous_tokens: Previously expanded tokens

        Returns:
            List of possible expansions

        Examples:
            >>> CommandShortcuts.get_possible_expansions('a', 0, [])
            ['announce']
            >>> CommandShortcuts.get_possible_expansions('a', 1, ['show'])
            ['adj-rib']
        """
        expansions = []
        for nickname, full_name, match_condition in cls.SHORTCUTS:
            is_match = token == nickname or token == full_name or (len(token) <= 2 and full_name.startswith(token))
            if is_match and match_condition(position, previous_tokens):
                if full_name not in expansions:
                    expansions.append(full_name)
        return expansions

    @classmethod
    def transform_cli_to_api(cls, command: str) -> str:
        """Transform CLI-friendly syntax to API-compatible syntax.

        Transforms:
        - 'neighbor <ip> show ...' → 'show neighbor <ip> ...'
        - 'adj-rib <in|out> show ...' → 'show adj-rib <in|out> ...'

        The CLI accepts more intuitive command-first syntax, but the API
        expects 'show'-first syntax. This transformation maintains API compatibility
        while improving CLI usability.

        Args:
            command: The command string after shortcut expansion

        Returns:
            Transformed command string

        Examples:
            >>> CommandShortcuts.transform_cli_to_api('neighbor 192.168.1.1 show summary')
            'show neighbor 192.168.1.1 summary'
            >>> CommandShortcuts.transform_cli_to_api('adj-rib in show extensive')
            'show adj-rib in extensive'
            >>> CommandShortcuts.transform_cli_to_api('show neighbor summary')
            'show neighbor summary'
        """
        tokens = command.split()
        if not tokens:
            return command

        # Pattern 1 (most specific): neighbor <ip> adj-rib <in|out> show [options...]
        # Transform to: show adj-rib <in|out> <ip> [options...]
        # Check this FIRST before general "neighbor <ip> show" pattern
        if len(tokens) >= 5 and tokens[0] == 'neighbor' and tokens[2] == 'adj-rib' and tokens[3] in ('in', 'out'):
            try:
                show_idx = tokens.index('show')
                if show_idx == 4:  # Must be 'neighbor <ip> adj-rib <in|out> show'
                    # Extract parts: ['neighbor', <ip>, 'adj-rib', 'in'/'out', 'show', [options...]]
                    neighbor_ip = tokens[1]
                    direction = tokens[3]  # 'in' or 'out'
                    options = tokens[show_idx + 1 :]  # Everything after 'show'

                    # Rebuild: 'show adj-rib <in|out> <ip> [options...]'
                    result = ['show', 'adj-rib', direction, neighbor_ip] + options
                    return ' '.join(result)
            except ValueError:
                # 'show' not in tokens, return unchanged
                pass

        # Pattern 2: neighbor <ip> [filters...] show [options...]
        # Transform to: show neighbor <ip> [filters...] [options...]
        if len(tokens) >= 3 and tokens[0] == 'neighbor':
            try:
                show_idx = tokens.index('show')
                if show_idx >= 2:  # Must have at least 'neighbor <ip> show'
                    # Extract parts: ['neighbor', <ip>, [filters...], 'show', [options...]]
                    neighbor_and_filters = tokens[1:show_idx]  # <ip> and any filters
                    options = tokens[show_idx + 1 :]  # Everything after 'show'

                    # Rebuild: 'show neighbor <ip> [filters...] [options...]'
                    result = ['show', 'neighbor'] + neighbor_and_filters + options
                    return ' '.join(result)
            except ValueError:
                # 'show' not in tokens, return unchanged
                pass

        # Pattern 3: adj-rib <in|out> show [options...]
        # Transform to: show adj-rib <in|out> [options...]
        if len(tokens) >= 3 and tokens[0] == 'adj-rib' and tokens[1] in ('in', 'out'):
            try:
                show_idx = tokens.index('show')
                if show_idx == 2:  # Must be 'adj-rib <in|out> show'
                    # Extract parts: ['adj-rib', 'in'/'out', 'show', [options...]]
                    direction = tokens[1]  # 'in' or 'out'
                    options = tokens[show_idx + 1 :]  # Everything after 'show'

                    # Rebuild: 'show adj-rib <in|out> [options...]'
                    result = ['show', 'adj-rib', direction] + options
                    return ' '.join(result)
            except ValueError:
                # 'show' not in tokens, return unchanged
                pass

        return command


# Convenience function for backward compatibility
def expand_shortcuts(command: str) -> str:
    """Expand shortcuts in a command string.

    Args:
        command: The command string to process

    Returns:
        Expanded command string
    """
    return CommandShortcuts.expand_shortcuts(command)



"""sr_policy/priority.py

SR Policy Priority Sub-TLV (type 15, RFC 9256 Section 2.4.6).

Wire format:
 +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 | Priority (1 octet)            |
 | Reserved (1 octet)            |
 +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
Total value length: 2 bytes.
"""

from __future__ import annotations

from struct import pack
from typing import ClassVar

from exabgp.bgp.message.update.attribute.tunnel_encap.tlv import SubTLV
from exabgp.util.types import Buffer


@SubTLV.register(15)
class PrioritySubTLV(SubTLV):
    """SR Policy Priority Sub-TLV."""

    SUBTYPE: ClassVar[int] = 15

    def __init__(self, priority: int) -> None:
        self.priority = priority

    def pack_value(self) -> bytes:
        return pack('!BB', self.priority, 0)

    def json(self) -> str:
        return f'"priority": {self.priority}'

    def __str__(self) -> str:
        return f'priority {self.priority}'

    @classmethod
    def unpack(cls, data: Buffer) -> PrioritySubTLV:
        priority = data[0] if data else 0
        return cls(priority=priority)



"""version.py

Created by Thomas Mangin on 2012-07-17.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from __future__ import annotations

from typing import ClassVar


class Version(int):
    BGP_4: ClassVar[int] = 4  # RFC 4271 - BGP version 4

    def pack_version(self) -> bytes:
        return bytes([self])



from __future__ import annotations

from typing import ClassVar

from exabgp.bgp.message.notification import Notify
from exabgp.bgp.message.update.nlri.mvpn.nlri import MVPN
from exabgp.bgp.message.update.nlri.qualifier import RouteDistinguisher
from exabgp.protocol.family import AFI
from exabgp.protocol.ip import IP, IPv4, IPv6
from exabgp.util.types import Buffer

# +-----------------------------------+
# |      RD   (8 octets)              |
# +-----------------------------------+
# | Multicast Source Length (1 octet) |
# +-----------------------------------+
# |   Multicast Source (variable)     |
# +-----------------------------------+
# |  Multicast Group Length (1 octet) |
# +-----------------------------------+
# |  Multicast Group (variable)       |
# +-----------------------------------+

# MVPN Source Active A-D Route length constants (RFC 6514)
MVPN_SOURCEAD_IPV4_LENGTH: int = 18  # 8 (RD) + 1 (source len) + 4 (IPv4) + 1 (group len) + 4 (IPv4)
MVPN_SOURCEAD_IPV6_LENGTH: int = 42  # 8 (RD) + 1 (source len) + 16 (IPv6) + 1 (group len) + 16 (IPv6)


@MVPN.register_mvpn(code=5)
class SourceAD(MVPN):
    NAME: ClassVar[str] = 'Source Active A-D Route'
    SHORT_NAME: ClassVar[str] = 'SourceAD'

    # Wire format offsets (after 2-byte type+length header)
    HEADER_SIZE = 2  # type(1) + length(1)

    def __init__(self, packed: Buffer, afi: AFI) -> None:
        """Create SourceAD from complete wire format bytes.

        Args:
            packed: Complete wire format (type + length + payload)
            afi: Address Family Identifier
        """
        MVPN.__init__(self, afi=afi)
        self._packed = bytes(packed)  # Ensure bytes for storage

    @classmethod
    def make_sourcead(
        cls,
        rd: RouteDistinguisher,
        afi: AFI,
        source: IP,
        group: IP,
    ) -> 'SourceAD':
        """Factory method to create SourceAD from semantic parameters."""
        payload = (
            bytes(rd.pack_rd())
            + bytes([len(source) * 8])
            + source.pack_ip()
            + bytes([len(group) * 8])
            + group.pack_ip()
        )
        # Prepend type + length header
        packed = bytes([cls.CODE, len(payload)]) + payload
        return cls(packed, afi)

    @property
    def rd(self) -> RouteDistinguisher:
        return RouteDistinguisher.unpack_routedistinguisher(self._packed[2:10])

    @property
    def source(self) -> IP:
        sourceiplen = int(self._packed[10] / 8)
        return IP.create_ip(self._packed[11 : 11 + sourceiplen])

    @property
    def group(self) -> IP:
        sourceiplen = int(self._packed[10] / 8)
        cursor = 11 + sourceiplen
        groupiplen = int(self._packed[cursor] / 8)
        return IP.create_ip(self._packed[cursor + 1 : cursor + 1 + groupiplen])

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, SourceAD)
            and self.CODE == other.CODE
            and self.rd == other.rd
            and self.source == other.source
            and self.group == other.group
        )

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __str__(self) -> str:
        return f'{self._prefix()}:{self.rd._str()}:{self.source!s}:{self.group!s}'

    def __hash__(self) -> int:
        # Direct _packed hash - include afi since MVPN supports both IPv4 and IPv6
        return hash((self.afi, self._packed))

    @classmethod
    def unpack_mvpn(cls, packed: Buffer, afi: AFI) -> 'MVPN':
        """Unpack SourceAD from complete wire format bytes.

        Args:
            packed: Complete wire format (type + length + payload)
            afi: Address Family Identifier
        """
        # packed includes header, payload starts at offset 2
        datalen = len(packed) - cls.HEADER_SIZE
        if datalen not in (MVPN_SOURCEAD_IPV4_LENGTH, MVPN_SOURCEAD_IPV6_LENGTH):  # IPv4 or IPv6
            raise Notify(3, 5, f'Unsupported Source Active A-D route length ({datalen} bytes).')

        # Validate source IP length (offset +2 for header)
        cursor = 10  # 2 (header) + 8 (RD)
        sourceiplen = int(packed[cursor] / 8)
        cursor += 1
        if sourceiplen != IPv4.BYTES and sourceiplen != IPv6.BYTES:
            raise Notify(
                3,
                5,
                f'Unsupported Source Active A-D Route Multicast Source IP length ({sourceiplen * 8} bits). Expected 32 bits (IPv4) or 128 bits (IPv6).',
            )
        cursor += sourceiplen

        # Validate group IP length
        groupiplen = int(packed[cursor] / 8)
        if groupiplen != IPv4.BYTES and groupiplen != IPv6.BYTES:
            raise Notify(
                3,
                5,
                f'Unsupported Source Active A-D Route Multicast Group IP length ({groupiplen * 8} bits). Expected 32 bits (IPv4) or 128 bits (IPv6).',
            )

        # Missing implementation of this check from RFC 6514:
        # Source Active A-D routes with a Multicast group belonging to the
        # Source Specific Multicast (SSM) range (as defined in [RFC4607], and
        # potentially extended locally on a router) MUST NOT be advertised by a
        # router and MUST be discarded if received.

        return cls(packed, afi)

    def json(self, announced: bool = True, compact: bool | None = None) -> str:
        content = ' "code": %d, ' % self.CODE
        content += '"parsed": true, '
        content += '"raw": "{}", '.format(self._raw())
        content += '"name": "{}", '.format(self.NAME)
        content += '{}, '.format(self.rd.json())
        content += '"source": "{}", '.format(str(self.source))
        content += '"group": "{}"'.format(str(self.group))
        return '{{{}}}'.format(content)



"""announce/label.py

Created by Thomas Mangin on 2017-07-05.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from __future__ import annotations

from exabgp.rib.route import Route

from exabgp.protocol.family import AFI
from exabgp.protocol.family import SAFI

from exabgp.bgp.message.update.nlri.label import Label
from exabgp.bgp.message.update.nlri.settings import INETSettings

from exabgp.configuration.announce import ParseAnnounce
from exabgp.configuration.announce.path import AnnouncePath
from exabgp.configuration.announce.route_builder import _build_route
from exabgp.configuration.core import Parser
from exabgp.configuration.core import Tokeniser
from exabgp.configuration.core import Scope
from exabgp.configuration.core import Error
from exabgp.configuration.schema import RouteBuilder, Leaf, ValueType, ActionTarget, ActionOperation, ActionKey
from exabgp.configuration.validator import LegacyParserValidator

from exabgp.configuration.static.parser import prefix
from exabgp.configuration.static.mpls import label


class AnnounceLabel(AnnouncePath):
    # Schema extends AnnouncePath with label using RouteBuilder (Settings mode)
    schema = RouteBuilder(
        description='MPLS labeled route announcement',
        nlri_class=Label,
        settings_class=INETSettings,
        prefix_parser=prefix,
        assign={
            **AnnouncePath.schema.assign,
            'label': 'labels',
        },
        children={
            **AnnouncePath.schema.children,
            'label': Leaf(
                type=ValueType.LABEL,
                description='MPLS label stack',
                target=ActionTarget.NLRI,
                operation=ActionOperation.SET,
                key=ActionKey.FIELD,
                validator=LegacyParserValidator(parser_func=label, name='label'),
            ),
        },
    )

    name = 'vpn'
    afi: AFI | None = None

    @property
    def syntax(self) -> str:
        """Syntax generated from schema."""
        return self.schema.syntax

    def __init__(self, parser: Parser, scope: Scope, error: Error) -> None:
        AnnouncePath.__init__(self, parser, scope, error)

    def clear(self) -> None:
        pass


@ParseAnnounce.register_family(AFI.ipv4, SAFI.nlri_mpls, ActionTarget.SCOPE, ActionOperation.EXTEND, ActionKey.NAME)
def nlri_mpls_v4(tokeniser: Tokeniser) -> list[Route]:
    return _build_route(tokeniser, AnnounceLabel.schema, AFI.ipv4, SAFI.nlri_mpls)


@ParseAnnounce.register_family(AFI.ipv6, SAFI.nlri_mpls, ActionTarget.SCOPE, ActionOperation.EXTEND, ActionKey.NAME)
def nlri_mpls_v6(tokeniser: Tokeniser) -> list[Route]:
    return _build_route(tokeniser, AnnounceLabel.schema, AFI.ipv6, SAFI.nlri_mpls)



"""vpls.py

Created by Nikita Shirokov on 2014-06-16.
Copyright (c) 2014-2017 Nikita Shirokov. All rights reserved.
Copyright (c) 2014-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from __future__ import annotations

from struct import pack, unpack
from typing import TYPE_CHECKING, Any

from exabgp.util.types import Buffer

if TYPE_CHECKING:
    from exabgp.bgp.message.open.capability.negotiated import Negotiated
    from exabgp.bgp.message.update.nlri.settings import VPLSSettings

from exabgp.bgp.message.action import Action
from exabgp.bgp.message.notification import Notify
from exabgp.bgp.message.update.nlri.nlri import NLRI
from exabgp.bgp.message.update.nlri.qualifier import RouteDistinguisher
from exabgp.bgp.message.update.nlri.qualifier.path import PathInfo
from exabgp.protocol.family import AFI, SAFI, Family


@NLRI.register(AFI.l2vpn, SAFI.vpls)
class VPLS(NLRI):
    """VPLS NLRI using packed-bytes-first pattern.

    _packed stores wire format:
    [length(2)][RD(8)][endpoint(2)][offset(2)][size(2)][base(3)] = 19 bytes

    AFI/SAFI set via Family parent class in __init__.

    Factory methods:
    - make_vpls(): Create from components (packs to wire format)
    - from_settings(): Create from VPLSSettings (validates before creation)
    - unpack_nlri(): Create from wire bytes (network receive path)
    """

    __slots__ = ()

    # Wire format length (including 2-byte length prefix)
    PACKED_LENGTH = 19  # length(2) + RD(8) + endpoint(2) + offset(2) + size(2) + base(3)

    def __init__(self, packed: Buffer) -> None:
        """Create a VPLS NLRI from packed wire-format bytes.

        Args:
            packed: 19 bytes: [length(2)][RD(8)][endpoint(2)][offset(2)][size(2)][base(3)]

        Note: action defaults to UNSET, set after creation (announce/withdraw).
        """
        NLRI.__init__(self, AFI.l2vpn, SAFI.vpls)
        self._packed: Buffer = bytes(packed)  # Ensure bytes for storage

    @classmethod
    def make_vpls(
        cls,
        rd: RouteDistinguisher,
        endpoint: int,
        base: int,
        offset: int,
        size: int,
        addpath: PathInfo = PathInfo.DISABLED,
    ) -> 'VPLS':
        """Factory method to create a VPLS NLRI from components.

        Args:
            rd: Route Distinguisher
            endpoint: VPLS endpoint (VE ID)
            base: Label base
            offset: Label block offset
            size: Label block size
            addpath: ADD-PATH path identifier

        Returns:
            New VPLS instance with packed wire format
        """
        packed = (
            b'\x00\x11'  # length prefix (17)
            + bytes(rd.pack_rd())
            + pack('!HHH', endpoint, offset, size)
            + pack('!L', (base << 4) | 0x1)[1:]  # 3 bytes with BOS bit
        )
        instance = cls(packed)
        instance.addpath = addpath
        return instance

    @classmethod
    def from_settings(cls, settings: 'VPLSSettings') -> 'VPLS':
        """Create VPLS NLRI from validated settings.

        This factory method validates settings and creates an immutable VPLS
        instance. Use this for deferred construction where all values are
        collected during parsing, then validated and used to create the NLRI.

        Args:
            settings: VPLSSettings with all required fields set

        Returns:
            Immutable VPLS NLRI instance

        Raises:
            ValueError: If settings validation fails
        """
        error = settings.validate()
        if error:
            raise ValueError(error)

        # Delegate to make_vpls which creates packed bytes
        assert settings.rd is not None
        assert settings.endpoint is not None
        assert settings.base is not None
        assert settings.offset is not None
        assert settings.size is not None

        instance = cls.make_vpls(
            rd=settings.rd,
            endpoint=settings.endpoint,
            base=settings.base,
            offset=settings.offset,
            size=settings.size,
        )
        # Note: settings.nexthop is now passed to Route, not stored in NLRI
        return instance

    @property
    def rd(self) -> RouteDistinguisher:
        """Route Distinguisher - unpacked from wire bytes."""
        return RouteDistinguisher(self._packed[2:10])

    @property
    def endpoint(self) -> int:
        """VPLS endpoint (VE ID) - unpacked from wire bytes."""
        return int(unpack('!H', bytes(self._packed[10:12]))[0])

    @property
    def offset(self) -> int:
        """Label block offset - unpacked from wire bytes."""
        return int(unpack('!H', bytes(self._packed[12:14]))[0])

    @property
    def block_size(self) -> int:
        """Label block size - unpacked from wire bytes."""
        return int(unpack('!H', bytes(self._packed[14:16]))[0])

    @property
    def base(self) -> int:
        """Label base - unpacked from wire bytes."""
        return int(unpack('!L', b'\x00' + bytes(self._packed[16:19]))[0]) >> 4

    def feedback(self, action: Action) -> str:
        """Validate VPLS NLRI-specific constraints.

        Note: nexthop validation is handled by Route.feedback(), not here.
        """
        # Size consistency check (for routes received from wire or created with invalid values)
        if self.base > (0xFFFFF - self.block_size):  # 20 bits, 3 bytes
            return 'vpls nlri size inconsistency'
        return ''

    def pack_nlri(self, negotiated: Negotiated) -> Buffer:
        # RFC 7911 ADD-PATH is possible for VPLS but not yet implemented
        # TODO: implement addpath support when negotiated.addpath.send(AFI.l2vpn, SAFI.vpls)
        return self._packed

    def index(self) -> bytes:
        return Family.index(self) + self._packed

    def json(self, announced: bool = True, compact: bool | None = None) -> str:
        # Note: The unique key for VPLS is the combination of all fields (rd, endpoint, base, offset, size).
        # This matches what index() returns for UPDATE withdraw matching.
        content = ', '.join(
            [
                self.rd.json(),
                '"endpoint": {}'.format(self.endpoint),
                '"base": {}'.format(self.base),
                '"offset": {}'.format(self.offset),
                '"size": {}'.format(self.block_size),
            ],
        )
        return '{{ {} }}'.format(content)

    def extensive(self) -> str:
        return 'vpls{} endpoint {} base {} offset {} size {}'.format(
            self.rd,
            self.endpoint,
            self.base,
            self.offset,
            self.block_size,
        )

    def __str__(self) -> str:
        return self.extensive()

    def __copy__(self) -> 'VPLS':
        new = self.__class__.__new__(self.__class__)
        # Family/NLRI slots - _packed is in NLRI slots
        self._copy_nlri_slots(new)
        return new

    def __deepcopy__(self, memo: dict[Any, Any]) -> 'VPLS':
        new = self.__class__.__new__(self.__class__)
        memo[id(self)] = new
        # Family/NLRI slots - _packed is in NLRI slots and is immutable bytes
        self._deepcopy_nlri_slots(new, memo)
        return new

    @classmethod
    def unpack_nlri(
        cls, afi: AFI, safi: SAFI, data: Buffer, action: Action, addpath: Any, negotiated: Negotiated
    ) -> tuple[VPLS, Buffer]:
        # Wire format: length(2) + RD(8) + endpoint(2) + offset(2) + size(2) + base(3) = 19 bytes
        if len(data) < 2:
            raise Notify(3, 10, f'VPLS NLRI too short: need at least 2 bytes, got {len(data)}')
        (length,) = unpack('!H', bytes(data[0:2]))
        if len(data) != length + 2:
            raise Notify(3, 10, 'l2vpn vpls message length is not consistent with encoded bgp')

        # Store wire format directly
        packed = bytes(data[0 : 2 + length])
        nlri = cls(packed)
        return nlri, data[2 + length :]



"""srlg.py

Created by Evelio Vila on 2016-12-01.
Copyright (c) 2014-2017 Exa Networks. All rights reserved.
"""

from __future__ import annotations

from struct import pack, unpack
from typing import Sequence
from exabgp.util import split

from exabgp.bgp.message.notification import Notify

from exabgp.bgp.message.update.attribute.bgpls.linkstate import LinkState
from exabgp.bgp.message.update.attribute.bgpls.linkstate import BaseLS
from exabgp.util.types import Buffer


#      0                   1                   2                   3
#      0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
#     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#     |              Type             |             Length            |
#     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#     |                  Shared Risk Link Group Value                 |
#     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#     //                         ............                        //
#     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#     |                  Shared Risk Link Group Value                 |
#     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#     https://tools.ietf.org/html/rfc7752#section-3.3.2.5 Shared Risk Link Group TLV


@LinkState.register_lsid(tlv=1096, json_key='shared-risk-link-groups', repr_name='link SRLG values')
class Srlg(BaseLS):
    @property
    def content(self) -> list[int]:
        """Unpack and return the SRLG values from packed bytes."""
        return [unpack('!L', chunk)[0] for chunk in split(self._packed, 4)]

    @classmethod
    def make_srlg(cls, srlg_values: Sequence[int]) -> Srlg:
        """Factory method to create Srlg from list of SRLG values."""
        packed = b''.join(pack('!I', v) for v in srlg_values)
        return cls(packed)

    @classmethod
    def unpack_bgpls(cls, data: Buffer) -> Srlg:
        if len(data) % 4:
            raise Notify(3, 5, 'Unable to decode SRLG')
        return cls(data)



"""line/rib.py

Created by Thomas Mangin on 2017-07-01.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from __future__ import annotations

import asyncio
import json
from typing import TYPE_CHECKING, Any

from exabgp.bgp.message.update.nlri.evpn.nlri import EVPN
from exabgp.bgp.message.update.nlri.flow import Flow
from exabgp.bgp.message.update.nlri.inet import INET
from exabgp.bgp.message.update.nlri.nlri import NLRI
from exabgp.bgp.message.update.nlri.vpls import VPLS
from exabgp.bgp.neighbor import NeighborTemplate
from exabgp.environment import getenv

if TYPE_CHECKING:
    from exabgp.reactor.api import API
    from exabgp.reactor.loop import Reactor


def register_rib() -> None:
    pass


def _show_adjrib_callback(
    reactor: 'Reactor',
    service: str,
    last: str,
    route_type: tuple[type[NLRI], ...],
    advertised: bool,
    rib_name: str,
    extensive: bool,
    use_json: bool,
) -> Any:
    def to_text(key: str, changes: list[Any]) -> None:
        for change in changes:
            if not isinstance(change.nlri, route_type):
                # log something about this drop?
                continue

            neighbor = reactor.neighbor_name(key) if extensive else reactor.neighbor_ip(key)
            family = f'{change.nlri.family().afi_safi()[0]} {change.nlri.family().afi_safi()[1]}'
            details = change.extensive() if extensive else str(change.nlri)
            msg = f'{neighbor} {family} {details}'
            reactor.processes.write(service, msg)

    def to_json(key: str, changes: list[Any]) -> None:
        jason: dict[str, dict[str, Any]] = {}
        neighbor = reactor.neighbor(key)
        neighbor_ip = reactor.neighbor_ip(key)
        routes = jason.setdefault(neighbor_ip, {'routes': []})['routes']

        if extensive and neighbor is not None:
            jason[neighbor_ip].update(NeighborTemplate.formated_dict(reactor.neighbor_cli_data(key)))

        for change in changes:
            if not isinstance(change.nlri, route_type):
                # log something about this drop?
                continue

            # After isinstance check, change.nlri is one of the route_type(s)
            nlri: Any = change.nlri
            route_entry: dict[str, Any] = {
                'family': str(change.nlri.family()).strip('()').replace(',', ''),
            }
            if hasattr(nlri, 'cidr'):
                route_entry['prefix'] = str(nlri.cidr.prefix())
            else:
                route_entry['nlri'] = str(nlri)
            routes.append(route_entry)

            for line in json.dumps(jason).split('\n'):
                reactor.processes.write(service, line)

    async def callback() -> None:
        lines_per_yield = getenv().api.chunk
        if last in ('routes', 'extensive', 'static', 'flow', 'l2vpn'):
            peers = reactor.peers()
        else:
            peers = [n for n in reactor.peers() if f'neighbor {last}' in n]
        for key in peers:
            routes = reactor.neighor_rib(key, rib_name, advertised)
            while routes:
                extracted, routes = routes[:lines_per_yield], routes[lines_per_yield:]
                if use_json:
                    to_json(key, extracted)
                else:
                    to_text(key, extracted)
                await asyncio.sleep(0)  # Yield control after each chunk (matches original yield True)
        await reactor.processes.answer_done(service)

    return callback


def show_adj_rib(self: 'API', reactor: 'Reactor', service: str, peers: list[str], command: str, use_json: bool) -> bool:
    # command contains direction and options (e.g., "in extensive" or "out")
    words = command.split()
    extensive = 'extensive' in words

    # Get direction from first word
    rib = words[0] if words else ''
    if rib not in ('in', 'out'):
        reactor.processes.answer_error_sync(service)
        return False

    klass: tuple[type[NLRI], ...] = (NLRI,)

    if 'inet' in words:
        klass = (INET,)
    elif 'flow' in words:
        klass = (Flow,)
    elif 'l2vpn' in words:
        klass = (VPLS, EVPN)

    if 'json' in words:
        words.remove('json')
        use_json = True

    for remove in ('in', 'out', 'extensive', 'inet', 'flow', 'l2vpn'):
        if remove in words:
            words.remove(remove)
    last = '' if not words else words[0]
    callback = _show_adjrib_callback(reactor, service, last, klass, False, rib, extensive, use_json)
    reactor.asynchronous.schedule(service, command, callback())
    return True


def flush_adj_rib_out(
    self: 'API', reactor: 'Reactor', service: str, peers: list[str], command: str, use_json: bool
) -> bool:
    async def callback(self: 'API', peer_list: list[str]) -> None:
        peer_names = ', '.join(peer_list) if peer_list else 'all peers'
        self.log_message(f'flushing adjb-rib out for {peer_names}')
        for peer_name in peer_list:
            reactor.neighbor_rib_resend(peer_name)
            await asyncio.sleep(0)  # Yield control after each peer (matches original yield False)

        await reactor.processes.answer_done(service)

    try:
        # peers list already parsed by dispatcher
        if not peers:
            self.log_failure(f'no neighbor matching the command : {command}', 'warning')
            reactor.processes.answer_error_sync(service)
            return False
        reactor.asynchronous.schedule(service, command, callback(self, peers))
        return True
    except ValueError:
        self.log_failure('issue parsing the command')
        reactor.processes.answer_error_sync(service)
        return False
    except IndexError:
        self.log_failure('issue parsing the command')
        reactor.processes.answer_error_sync(service)
        return False


def clear_adj_rib(
    self: 'API', reactor: 'Reactor', service: str, peers: list[str], command: str, use_json: bool
) -> bool:
    async def callback(self: 'API', peer_list: list[str], direction: str) -> None:
        peer_names = ', '.join(peer_list) if peer_list else 'all peers'
        self.log_message(f'clearing adjb-rib-{direction} for {peer_names}')
        for peer_name in peer_list:
            if direction == 'out':
                reactor.neighbor_rib_out_withdraw(peer_name)
            else:
                reactor.neighbor_rib_in_clear(peer_name)
            await asyncio.sleep(0)  # Yield control after each peer (matches original yield False)

        await reactor.processes.answer_done(service)

    try:
        # peers list already parsed by dispatcher
        if not peers:
            self.log_failure(f'no neighbor matching the command : {command}', 'warning')
            reactor.processes.answer_error_sync(service)
            return False
        words = command.split()
        direction = 'in' if 'in' in words else 'out'
        reactor.asynchronous.schedule(service, command, callback(self, peers, direction))
        return True
    except ValueError:
        self.log_failure('issue parsing the command')
        reactor.processes.answer_error_sync(service)
        return False
    except IndexError:
        self.log_failure('issue parsing the command')
        reactor.processes.answer_error_sync(service)
        return False



"""srv6/sidinformation.py

Created by Ryoga Saito 2022-02-24
Copyright (c) 2022 Ryoga Saito. All rights reserved.
"""

from __future__ import annotations

from struct import pack, unpack
from typing import Any, Callable, ClassVar, Protocol, Type, TypeVar

from exabgp.protocol.ip import IPv6

from exabgp.bgp.message.notification import Notify
from exabgp.bgp.message.update.attribute.sr.srv6.l2service import Srv6L2Service
from exabgp.bgp.message.update.attribute.sr.srv6.l3service import Srv6L3Service
from exabgp.bgp.message.update.attribute.sr.srv6.generic import GenericSrv6ServiceDataSubSubTlv
from exabgp.util.types import Buffer


class HasTLV(Protocol):
    """Protocol for classes with TLV class attribute and unpack_attribute method."""

    TLV: ClassVar[int]

    @classmethod
    def unpack_attribute(cls, data: Buffer, length: int) -> Any: ...


# TypeVar for classes with TLV attribute
SubSubTlvType = TypeVar('SubSubTlvType', bound=HasTLV)

# 3.1.  SRv6 SID Information Sub-TLV
#
#  0                   1                   2                   3
#  0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# | SRv6 Service  |    SRv6 Service               |               |
# | Sub-TLV       |    Sub-TLV                    |               |
# | Type=1        |    Length                     |  RESERVED1    |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |  SRv6 SID Value (16 octets)                                  //
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# | Svc SID Flags |   SRv6 Endpoint Behavior      |   RESERVED2   |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |  SRv6 Service Data Sub-Sub-TLVs                              //
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

#            Figure 3: SRv6 SID Information Sub-TLV


@Srv6L2Service.register()
@Srv6L3Service.register()
class Srv6SidInformation:
    TLV: ClassVar[int] = 1

    # Registry maps TLV codes to Sub-Sub-TLV classes (uses HasTLV protocol)
    registered_subsubtlvs: ClassVar[dict[int, Type[HasTLV]]] = dict()

    def __init__(
        self,
        sid: IPv6,
        behavior: int,
        subsubtlvs: list[Any],
        packed: Buffer | None = None,
    ) -> None:
        """Initialize SID Information sub-TLV.

        Args:
            sid: SRv6 SID value (IPv6 address)
            behavior: SRv6 Endpoint Behavior code
            subsubtlvs: List of sub-sub-TLVs (Srv6SidStructure or GenericSrv6ServiceDataSubSubTlv)
            packed: Optional pre-packed wire format
        """
        self.sid: IPv6 = sid
        self.behavior: int = behavior
        self.subsubtlvs: list[Any] = subsubtlvs
        self.packed: Buffer = self.pack_tlv()

    @classmethod
    def register(cls) -> Callable[[Type[SubSubTlvType]], Type[SubSubTlvType]]:
        def register_subsubtlv(klass: Type[SubSubTlvType]) -> Type[SubSubTlvType]:
            code: int = klass.TLV
            if code in cls.registered_subsubtlvs:
                raise RuntimeError('only one class can be registered per SRv6 Service Sub-Sub-TLV type')
            cls.registered_subsubtlvs[code] = klass
            return klass

        return register_subsubtlv

    @classmethod
    def unpack_attribute(cls, data: Buffer, length: int) -> Srv6SidInformation:
        # SRv6 SID Information: reserved(1) + SID(16) + flags(1) + behavior(2) + reserved(1) = 21 bytes minimum
        if len(data) < 21:
            raise Notify(3, 1, f'SRv6 SID Information too short: need 21 bytes, got {len(data)}')
        sid: IPv6 = IPv6.unpack_ipv6(data[1:17])
        behavior: int = unpack('!H', data[18:20])[0]
        subsubtlvs: list[GenericSrv6ServiceDataSubSubTlv] = []

        data = data[21:]
        while data:
            # Sub-Sub-TLV header: type(1) + length(2) = 3 bytes minimum
            if len(data) < 3:
                raise Notify(3, 1, f'SRv6 Sub-Sub-TLV header truncated: need 3 bytes, got {len(data)}')
            code: int = data[0]
            length = unpack('!H', data[1:3])[0]
            if len(data) < length + 3:
                raise Notify(3, 1, f'SRv6 Sub-Sub-TLV truncated: need {length + 3} bytes, got {len(data)}')
            if code in cls.registered_subsubtlvs:
                subsubtlv: GenericSrv6ServiceDataSubSubTlv = cls.registered_subsubtlvs[code].unpack_attribute(
                    data[3 : length + 3], length
                )
            else:
                subsubtlv = GenericSrv6ServiceDataSubSubTlv(data[3 : length + 3], code)
            subsubtlvs.append(subsubtlv)
            data = data[length + 3 :]

        return cls(sid=sid, behavior=behavior, subsubtlvs=subsubtlvs)

    def pack_tlv(self) -> bytes:
        subsubtlvs_packed: bytes = b''.join([_.pack_tlv() for _ in self.subsubtlvs])
        length: int = len(subsubtlvs_packed) + 21
        reserved: int = 0
        flags: int = 0

        return (
            pack('!B', self.TLV)
            + pack('!H', length)
            + pack('!B', reserved)
            + self.sid.pack_ip()
            + pack('!B', flags)
            + pack('!H', self.behavior)
            + pack('!B', reserved)
            + subsubtlvs_packed
        )

    def __str__(self) -> str:
        s: str = 'sid-information [ sid:{} flags:0 endpoint_behavior:0x{:x} '.format(str(self.sid), self.behavior)
        if len(self.subsubtlvs) != 0:
            s += ' [ ' + ', '.join([str(subsubtlv) for subsubtlv in self.subsubtlvs]) + ' ]'
        s + ' ]'
        return s

    def json(self, compact: bool | None = None) -> str:
        s: str = '{{ "sid": "{}", "flags": 0, "endpoint_behavior": {}'.format(str(self.sid), self.behavior)
        content: str = ', '.join(subsubtlv.json() for subsubtlv in self.subsubtlvs)
        if content:
            s += ', {}'.format(content)
        s += ' }'
        return s



"""message.py

Created by Thomas Mangin on 2010-01-15.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from __future__ import annotations

from struct import pack
from typing import TYPE_CHECKING, Callable, ClassVar, Type

from exabgp.util.types import Buffer

if TYPE_CHECKING:
    from exabgp.bgp.message.open.capability.negotiated import Negotiated


class _MessageCode(int):
    OPEN: ClassVar[int] = 0x01  # .          1
    UPDATE: ClassVar[int] = 0x02  # .        2
    NOTIFICATION: ClassVar[int] = 0x03  # .  3
    KEEPALIVE: ClassVar[int] = 0x04  # .     4
    ROUTE_REFRESH: ClassVar[int] = 0x05  # . 5
    OPERATIONAL: ClassVar[int] = 0x06  # .   6  # Not IANA assigned yet
    NOP: ClassVar[int] = 0xFC  # .           252 - internal - no data yet
    DONE: ClassVar[int] = 0xFD  # .          253 - internal - peer finished
    AWAKE: ClassVar[int] = 0xFE  # .         254 - internal - immediate action

    names: ClassVar[dict[int | None, str]] = {
        None: 'INVALID',
        NOP: 'NOP',
        AWAKE: 'AWAKE',
        DONE: 'DONE',
        OPEN: 'OPEN',
        UPDATE: 'UPDATE',
        NOTIFICATION: 'NOTIFICATION',
        KEEPALIVE: 'KEEPALIVE',
        ROUTE_REFRESH: 'ROUTE_REFRESH',
        OPERATIONAL: 'OPERATIONAL',
    }

    short_names: ClassVar[dict[int | None, str]] = {
        None: 'invalid',
        NOP: 'nop',
        AWAKE: 'awake',
        DONE: 'done',
        OPEN: 'open',
        UPDATE: 'update',
        NOTIFICATION: 'notification',
        KEEPALIVE: 'keepalive',
        ROUTE_REFRESH: 'refresh',
        OPERATIONAL: 'operational',
    }

    long_names: ClassVar[dict[int | None, str]] = {
        None: 'invalid',
        NOP: 'nop',
        AWAKE: 'awake',
        DONE: 'done',
        OPEN: 'open',
        UPDATE: 'update',
        NOTIFICATION: 'notification',
        KEEPALIVE: 'keepalive',
        ROUTE_REFRESH: 'route-refresh',
        OPERATIONAL: 'operational',
    }

    # to_short_names = dict((name,code) for (code,name) in short_names.items())

    SHORT: str
    NAME: str

    def __init__(self, value: int) -> None:
        self.SHORT = self.short()
        self.NAME = str(self)

    def __str__(self) -> str:
        return self.names.get(self, 'unknown message {}'.format(hex(self)))

    def __repr__(self) -> str:
        return str(self)

    def short(self) -> str:
        return self.short_names.get(self, '{}'.format(self))


# ================================================================== BGP Message
#

# 0                   1                   2                   3
# 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |                                                               |
# +                                                               +
# |                                                               |
# +                                                               +
# |                           Marker                              |
# +                                                               +
# |                                                               |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |          Length               |      Type     |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+


class Message:
    # we need to define TYPE inside __init__ of the subclasses
    # otherwise we can not dynamically create different UnknownMessage
    # TYPE = None

    MARKER: ClassVar[bytes] = bytes(
        [
            0xFF,
        ]
        * 16,
    )
    HEADER_LEN: ClassVar[int] = 19

    registered_message: ClassVar[dict[int, Type[Message]]] = {}
    klass_unknown: ClassVar[Callable[[int, Buffer, Negotiated], Message]]

    # TYPE attribute set by subclasses
    TYPE: ClassVar[bytes]
    ID: ClassVar[int]

    # Reactor scheduling - 0 (MESSAGE) for real BGP messages
    # Scheduling messages (NOP, AWAKE, DONE) set this to Scheduling.LATER/NOW/CLOSE
    SCHEDULING: int = 0  # Scheduling.MESSAGE

    class CODE:
        NOP: ClassVar[_MessageCode] = _MessageCode(_MessageCode.NOP)
        OPEN: ClassVar[_MessageCode] = _MessageCode(_MessageCode.OPEN)
        UPDATE: ClassVar[_MessageCode] = _MessageCode(_MessageCode.UPDATE)
        NOTIFICATION: ClassVar[_MessageCode] = _MessageCode(_MessageCode.NOTIFICATION)
        KEEPALIVE: ClassVar[_MessageCode] = _MessageCode(_MessageCode.KEEPALIVE)
        ROUTE_REFRESH: ClassVar[_MessageCode] = _MessageCode(_MessageCode.ROUTE_REFRESH)
        OPERATIONAL: ClassVar[_MessageCode] = _MessageCode(_MessageCode.OPERATIONAL)
        DONE: ClassVar[_MessageCode] = _MessageCode(_MessageCode.DONE)
        AWAKE: ClassVar[_MessageCode] = _MessageCode(_MessageCode.AWAKE)

        MESSAGES: ClassVar[list[_MessageCode]] = [
            NOP,
            OPEN,
            UPDATE,
            NOTIFICATION,
            KEEPALIVE,
            ROUTE_REFRESH,
            OPERATIONAL,
        ]

        @staticmethod
        def name(message_id: int | None) -> str:
            if message_id is None:
                return _MessageCode.names.get(message_id, 'unknown message')
            return _MessageCode.names.get(message_id, 'unknown message {}'.format(hex(message_id)))

        @staticmethod
        def short(message_id: int | None) -> str:
            if message_id is None:
                return _MessageCode.short_names.get(message_id, 'unknown message')
            return _MessageCode.short_names.get(message_id, 'unknown message {}'.format(hex(message_id)))

        # # Can raise KeyError
        # @staticmethod
        # def code (short):
        # 	return _MessageCode.names.get[short]

        def __init__(self) -> None:
            raise RuntimeError('This class can not be instantiated')

    Length: ClassVar[dict[int, Callable[[int], bool]]] = {
        CODE.OPEN: lambda _: _ >= 29,  # noqa
        CODE.UPDATE: lambda _: _ >= 23,  # noqa
        CODE.NOTIFICATION: lambda _: _ >= 21,  # noqa
        CODE.KEEPALIVE: lambda _: _ == 19,  # noqa
        CODE.ROUTE_REFRESH: lambda _: _ == 23,  # noqa
    }

    @staticmethod
    def string(code: int | None) -> str:
        return _MessageCode.long_names.get(code, 'unknown')

    def _message(self, message: Buffer) -> bytes:
        # Accept Buffer (bytes or memoryview), convert to bytes for output
        message_len: bytes = pack('!H', 19 + len(message))
        return self.MARKER + message_len + self.TYPE + bytes(message)

    def pack_message(self, negotiated: Negotiated) -> Buffer:
        raise NotImplementedError('message not implemented in subclasses')

    @classmethod
    def unpack_message(cls, data: Buffer, negotiated: Negotiated) -> Message:
        raise NotImplementedError('unpack_message not implemented in subclass')

    @classmethod
    def register(cls, klass: Type[Message]) -> Type[Message]:
        if klass.ID in cls.registered_message:
            raise RuntimeError('only one class can be registered per message')
        cls.registered_message[klass.ID] = klass
        return klass

    @classmethod
    def klass(cls, what: int) -> Type[Message]:
        if what in cls.registered_message:
            return cls.registered_message[what]
        from exabgp.bgp.message.notification import Notify

        raise Notify(2, 4, f'can not handle message {what}')

    @classmethod
    def unpack(cls, message: int, data: Buffer, negotiated: Negotiated) -> Message:
        """Unpack a BGP message from wire format.

        Args:
            message: BGP message type code
            data: Message body as Buffer (bytes, memoryview, etc. - PEP 688)
            negotiated: Negotiated capabilities for this session

        Returns:
            Parsed Message subclass instance
        """
        if message in cls.registered_message:
            return cls.klass(message).unpack_message(data, negotiated)
        return cls.klass_unknown(message, data, negotiated)

    @classmethod
    def code_from_name(cls, name: str) -> _MessageCode:
        for message in cls.CODE.MESSAGES:
            if name == str(message) or name == message.short():
                return message
        return cls.CODE.NOP



# http://teethgrinder.co.uk/perm.php?a=Python-memory-leak-detector
from __future__ import annotations

import gc
import inspect


def dump():
    # force collection
    print('\nCollecting GARBAGE:')
    gc.collect()
    # prove they have been collected
    print('\nCollecting GARBAGE:')
    gc.collect()

    print('\nGARBAGE OBJECTS:')
    for x in gc.garbage:
        s = str(x)
        if len(s) > 80:
            s = '%s...' % s[:80]

        print('::', s)
        print('		type:', type(x))
        print('   referrers:', len(gc.get_referrers(x)))
        try:
            print('	is class:', inspect.isclass(type(x)))
            print('	  module:', inspect.getmodule(x))

            lines, line_num = inspect.getsourcelines(type(x))
            print('	line num:', line_num)
            for line in lines:
                print('		line:', line.rstrip('\n'))
        except Exception:
            pass

        print()



"""sr_policy/binding_sid.py

SR Policy Binding SID Sub-TLV (type 13, RFC 9256 Section 2.4.2).

Wire format:
 +-+-+-+-+-+-+-+-+
 | Flags (1 octet)|
 +-+-+-+-+-+-+-+-+
 | Reserved (1 octet) |
 +-+-+-+-+-+-+-+-+
 | BSID (0 or 4 octets) |
 +-+-+-+-+-+-+-+-+

When no BSID is present (explicit null), value is 2 bytes (flags + reserved only).
When an MPLS label is present, value is 6 bytes: flags(1) + reserved(1) + label_entry(4).

MPLS Label Stack Entry (4 bytes):
  bits [31:12]: Label (top 20 bits)
  bits [11:9]:  TC (3 bits)
  bit  [8]:     S (bottom-of-stack)
  bits [7:0]:   TTL (8 bits)
"""

from __future__ import annotations

from struct import pack, unpack
from typing import ClassVar

from exabgp.bgp.message.update.attribute.tunnel_encap.tlv import SubTLV
from exabgp.util.types import Buffer

_BSID_FLAG_SPECIFIED = 0x10  # B flag: BSID is explicitly specified


@SubTLV.register(13)
class BindingSIDSubTLV(SubTLV):
    """SR Policy Binding SID Sub-TLV (MPLS)."""

    SUBTYPE: ClassVar[int] = 13

    def __init__(self, label: int | None = None, flags: int = 0) -> None:
        """Args:
        label: MPLS label value (top 20 bits of label stack entry), None = no BSID.
        flags: Sub-TLV flags byte.
        """
        self.label = label
        self.flags = flags

    def pack_value(self) -> bytes:
        if self.label is None:
            return pack('!BB', self.flags, 0)
        label_entry = (self.label << 12) | 0x100  # S=1, TTL=0
        return pack('!BBL', self.flags | _BSID_FLAG_SPECIFIED, 0, label_entry)

    def json(self) -> str:
        if self.label is None:
            return '"binding-sid": null'
        return f'"binding-sid": {{"type": "mpls", "label": {self.label}}}'

    def __str__(self) -> str:
        if self.label is None:
            return 'binding-sid null'
        return f'binding-sid mpls {self.label}'

    @classmethod
    def unpack(cls, data: Buffer) -> BindingSIDSubTLV:
        if len(data) < 2:
            return cls()
        flags = data[0]
        if len(data) >= 6:
            label_entry: int = unpack('!L', data[2:6])[0]
            label = label_entry >> 12
            return cls(label=label, flags=flags)
        return cls(label=None, flags=flags)



"""bgpls/__init__.py

BGP-LS (Link State) NLRI Implementation.

References:
    RFC 7752 - North-Bound Distribution of Link-State and TE Information Using BGP
               https://datatracker.ietf.org/doc/html/rfc7752
    RFC 9513 - BGP-LS Extensions for SRv6
               https://datatracker.ietf.org/doc/html/rfc9513

BGP-LS NLRI Format (RFC 7752 Section 3.2):
+-----------------------------------+
|    NLRI Type (2 octets)           |
+-----------------------------------+
|    Total NLRI Length (2 octets)   |
+-----------------------------------+
|    NLRI Value (variable)          |
+-----------------------------------+

NLRI Types (RFC 7752 Section 3.2.1):
- Type 1: Node NLRI - NODE
- Type 2: Link NLRI - LINK
- Type 3: IPv4 Topology Prefix NLRI - PREFIXv4
- Type 4: IPv6 Topology Prefix NLRI - PREFIXv6
- Type 6: SRv6 SID NLRI - SRv6SID (RFC 9513)

AFI/SAFI: BGP-LS (16388) / BGP-LS (71)

Wire Format Reference: doc/RFC_WIRE_FORMAT_REFERENCE.md#bgp-ls-rfc-7752--rfc-9552

Created by Evelio Vila on 2016-11-26. eveliovila@gmail.com
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

# Every BGP LINK_STATE NLRI should be imported from this file
# as it makes sure that all the registering decorator are run

# flake8: noqa: F401,E261

from __future__ import annotations

from exabgp.bgp.message.update.nlri.bgpls.nlri import BGPLS

from exabgp.bgp.message.update.nlri.bgpls.node import NODE
from exabgp.bgp.message.update.nlri.bgpls.link import LINK
from exabgp.bgp.message.update.nlri.bgpls.prefixv4 import PREFIXv4
from exabgp.bgp.message.update.nlri.bgpls.prefixv6 import PREFIXv6
from exabgp.bgp.message.update.nlri.bgpls.srv6sid import SRv6SID

__all__ = [
    'BGPLS',
    'NODE',
    'LINK',
    'PREFIXv4',
    'PREFIXv6',
    'SRv6SID',
]



"""match.py

Created by Thomas Mangin on 2015-06-22.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from __future__ import annotations

from typing import Any

from exabgp.configuration.core import Section
from exabgp.configuration.core import Parser
from exabgp.configuration.core import Scope
from exabgp.configuration.core import Error
from exabgp.configuration.schema import ActionKey, ActionOperation, ActionTarget, Container, Leaf, ValueType


from exabgp.configuration.flow.parser import source
from exabgp.configuration.flow.parser import destination
from exabgp.configuration.flow.parser import any_port
from exabgp.configuration.flow.parser import source_port
from exabgp.configuration.flow.parser import destination_port
from exabgp.configuration.flow.parser import tcp_flags
from exabgp.configuration.flow.parser import protocol
from exabgp.configuration.flow.parser import next_header
from exabgp.configuration.flow.parser import fragment
from exabgp.configuration.flow.parser import packet_length
from exabgp.configuration.flow.parser import icmp_code
from exabgp.configuration.flow.parser import icmp_type
from exabgp.configuration.flow.parser import dscp
from exabgp.configuration.flow.parser import traffic_class
from exabgp.configuration.flow.parser import flow_label


class ParseFlowMatch(Section):
    # Schema definition for FlowSpec match criteria
    # All match fields use NLRI target, APPEND operation, FIELD key
    schema = Container(
        description='FlowSpec match criteria',
        children={
            'source': Leaf(
                type=ValueType.IP_PREFIX,
                description='Source IP prefix to match',
                target=ActionTarget.NLRI,
                operation=ActionOperation.APPEND,
                key=ActionKey.FIELD,
            ),
            'source-ipv4': Leaf(
                type=ValueType.IP_PREFIX,
                description='Source IPv4 prefix to match',
                target=ActionTarget.NLRI,
                operation=ActionOperation.APPEND,
                key=ActionKey.FIELD,
            ),
            'source-ipv6': Leaf(
                type=ValueType.IP_PREFIX,
                description='Source IPv6 prefix to match',
                target=ActionTarget.NLRI,
                operation=ActionOperation.APPEND,
                key=ActionKey.FIELD,
            ),
            'destination': Leaf(
                type=ValueType.IP_PREFIX,
                description='Destination IP prefix to match',
                target=ActionTarget.NLRI,
                operation=ActionOperation.APPEND,
                key=ActionKey.FIELD,
            ),
            'destination-ipv4': Leaf(
                type=ValueType.IP_PREFIX,
                description='Destination IPv4 prefix to match',
                target=ActionTarget.NLRI,
                operation=ActionOperation.APPEND,
                key=ActionKey.FIELD,
            ),
            'destination-ipv6': Leaf(
                type=ValueType.IP_PREFIX,
                description='Destination IPv6 prefix to match',
                target=ActionTarget.NLRI,
                operation=ActionOperation.APPEND,
                key=ActionKey.FIELD,
            ),
            'protocol': Leaf(
                type=ValueType.STRING,
                description='IP protocol to match (IPv4 only)',
                target=ActionTarget.NLRI,
                operation=ActionOperation.APPEND,
                key=ActionKey.FIELD,
            ),
            'next-header': Leaf(
                type=ValueType.STRING,
                description='Next header to match (IPv6 only)',
                target=ActionTarget.NLRI,
                operation=ActionOperation.APPEND,
                key=ActionKey.FIELD,
            ),
            'port': Leaf(
                type=ValueType.STRING,
                description='Port range to match (source or destination)',
                target=ActionTarget.NLRI,
                operation=ActionOperation.APPEND,
                key=ActionKey.FIELD,
            ),
            'destination-port': Leaf(
                type=ValueType.STRING,
                description='Destination port range to match',
                target=ActionTarget.NLRI,
                operation=ActionOperation.APPEND,
                key=ActionKey.FIELD,
            ),
            'source-port': Leaf(
                type=ValueType.STRING,
                description='Source port range to match',
                target=ActionTarget.NLRI,
                operation=ActionOperation.APPEND,
                key=ActionKey.FIELD,
            ),
            'icmp-type': Leaf(
                type=ValueType.INTEGER,
                description='ICMP type to match (IPv6 only)',
                target=ActionTarget.NLRI,
                operation=ActionOperation.APPEND,
                key=ActionKey.FIELD,
            ),
            'icmp-code': Leaf(
                type=ValueType.INTEGER,
                description='ICMP code to match (IPv6 only)',
                target=ActionTarget.NLRI,
                operation=ActionOperation.APPEND,
                key=ActionKey.FIELD,
            ),
            'tcp-flags': Leaf(
                type=ValueType.STRING,
                description='TCP flags to match',
                target=ActionTarget.NLRI,
                operation=ActionOperation.APPEND,
                key=ActionKey.FIELD,
            ),
            'packet-length': Leaf(
                type=ValueType.STRING,
                description='Packet length range to match',
                target=ActionTarget.NLRI,
                operation=ActionOperation.APPEND,
                key=ActionKey.FIELD,
            ),
            'dscp': Leaf(
                type=ValueType.INTEGER,
                description='DSCP value to match',
                target=ActionTarget.NLRI,
                operation=ActionOperation.APPEND,
                key=ActionKey.FIELD,
            ),
            'traffic-class': Leaf(
                type=ValueType.INTEGER,
                description='Traffic class to match (IPv6 only)',
                target=ActionTarget.NLRI,
                operation=ActionOperation.APPEND,
                key=ActionKey.FIELD,
            ),
            'fragment': Leaf(
                type=ValueType.STRING,
                description='Fragment flags to match (IPv4 only)',
                target=ActionTarget.NLRI,
                operation=ActionOperation.APPEND,
                key=ActionKey.FIELD,
            ),
            'flow-label': Leaf(
                type=ValueType.STRING,
                description='Flow label to match (IPv6 only)',
                target=ActionTarget.NLRI,
                operation=ActionOperation.APPEND,
                key=ActionKey.FIELD,
            ),
        },
    )
    definition: list[str] = [
        'source 10.0.0.0/24',
        'source ::1/128/0',
        'destination 10.0.1.0/24',
        'port 25',
        'source-port >1024',
        'destination-port [ =80 =3128 >8080&<8088 ]',
        'packet-length [ >200&<300 >400&<500 ]',
        'tcp-flags [ 0x20+0x8+0x1 #name-here ]  # to check',
        '(ipv4 only) protocol [ udp tcp ]',
        '(ipv4 only) fragment [ dont-fragment is-fragment first-fragment last-fragment ]',
        '(ipv6 only) next-header [ udp tcp ]',
        '(ipv6 only) flow-label >100&<2000',
        '(ipv6 only) icmp-type 35  # to check',
        '(ipv6 only) icmp-code 55  # to check',
    ]

    joined: str = ';\\n  '.join(definition)
    syntax: str = f'match {{\n  {joined};\n}}'

    known: dict[str | tuple[Any, ...], object] = {
        'source': source,
        'source-ipv4': source,
        'source-ipv6': source,
        'destination': destination,
        'destination-ipv4': destination,
        'destination-ipv6': destination,
        'protocol': protocol,
        'next-header': next_header,
        'port': any_port,
        'destination-port': destination_port,
        'source-port': source_port,
        'icmp-type': icmp_type,
        'icmp-code': icmp_code,
        'tcp-flags': tcp_flags,
        'packet-length': packet_length,
        'dscp': dscp,
        'traffic-class': traffic_class,
        'fragment': fragment,
        'flow-label': flow_label,
    }

    # action dict removed - derived from schema

    name: str = 'flow/match'

    def __init__(self, parser: Parser, scope: Scope, error: Error) -> None:
        Section.__init__(self, parser, scope, error)

    def clear(self) -> None:
        pass

    def pre(self) -> bool:
        return True

    def post(self) -> bool:
        return True



from __future__ import annotations

from struct import pack
from typing import ClassVar

from exabgp.bgp.message.notification import Notify
from exabgp.bgp.message.update.nlri.mvpn.nlri import MVPN
from exabgp.bgp.message.update.nlri.qualifier import RouteDistinguisher
from exabgp.protocol.family import AFI
from exabgp.protocol.ip import IP, IPv4, IPv6
from exabgp.util.types import Buffer

# +-----------------------------------+
# |      RD   (8 octets)              |
# +-----------------------------------+
# |    Source AS (4 octets)           |
# +-----------------------------------+
# | Multicast Source Length (1 octet) |
# +-----------------------------------+
# |   Multicast Source (variable)     |
# +-----------------------------------+
# |  Multicast Group Length (1 octet) |
# +-----------------------------------+
# |  Multicast Group   (variable)     |
# +-----------------------------------+

# MVPN Shared Join Route length constants (RFC 6514)
MVPN_SHAREDJOIN_IPV4_LENGTH: int = 22  # 8 (RD) + 4 (Source AS) + 1 (source len) + 4 (IPv4) + 1 (group len) + 4 (IPv4)
MVPN_SHAREDJOIN_IPV6_LENGTH: int = 46  # 8 (RD) + 4 (Source AS) + 1 (source len) + 16 (IPv6) + 1 (group len) + 16 (IPv6)


@MVPN.register_mvpn(code=6)
class SharedJoin(MVPN):
    NAME: ClassVar[str] = 'C-Multicast Shared Tree Join route'
    SHORT_NAME: ClassVar[str] = 'Shared-Join'

    # Wire format offsets (after 2-byte type+length header)
    HEADER_SIZE = 2  # type(1) + length(1)

    def __init__(self, packed: Buffer, afi: AFI) -> None:
        """Create SharedJoin from complete wire format bytes.

        Args:
            packed: Complete wire format (type + length + payload)
            afi: Address Family Identifier
        """
        MVPN.__init__(self, afi=afi)
        self._packed = bytes(packed)  # Ensure bytes for storage

    @classmethod
    def make_sharedjoin(
        cls,
        rd: RouteDistinguisher,
        afi: AFI,
        source: IP,
        group: IP,
        source_as: int,
    ) -> 'SharedJoin':
        """Factory method to create SharedJoin from semantic parameters."""
        payload = (
            bytes(rd.pack_rd())
            + pack('!I', source_as)
            + bytes([len(source) * 8])
            + source.pack_ip()
            + bytes([len(group) * 8])
            + group.pack_ip()
        )
        # Prepend type + length header
        packed = bytes([cls.CODE, len(payload)]) + payload
        return cls(packed, afi)

    @property
    def rd(self) -> RouteDistinguisher:
        return RouteDistinguisher.unpack_routedistinguisher(self._packed[2:10])

    @property
    def source_as(self) -> int:
        return int.from_bytes(self._packed[10:14], 'big')

    @property
    def source(self) -> IP:
        cursor = 14  # 2 (header) + 8 (RD) + 4 (source_as)
        sourceiplen = int(self._packed[cursor] / 8)
        return IP.create_ip(self._packed[cursor + 1 : cursor + 1 + sourceiplen])

    @property
    def group(self) -> IP:
        cursor = 14  # 2 (header) + 8 (RD) + 4 (source_as)
        sourceiplen = int(self._packed[cursor] / 8)
        cursor += 1 + sourceiplen
        groupiplen = int(self._packed[cursor] / 8)
        return IP.create_ip(self._packed[cursor + 1 : cursor + 1 + groupiplen])

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, SharedJoin)
            and self.CODE == other.CODE
            and self.rd == other.rd
            and self.source == other.source
            and self.group == other.group
        )

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __str__(self) -> str:
        return f'{self._prefix()}:{self.rd._str()}:{self.source_as!s}:{self.source!s}:{self.group!s}'

    def __hash__(self) -> int:
        # Direct _packed hash - include afi since MVPN supports both IPv4 and IPv6
        return hash((self.afi, self._packed))

    @classmethod
    def unpack_mvpn(cls, packed: Buffer, afi: AFI) -> 'MVPN':
        """Unpack SharedJoin from complete wire format bytes.

        Args:
            packed: Complete wire format (type + length + payload)
            afi: Address Family Identifier
        """
        # packed includes header, payload starts at offset 2
        datalen = len(packed) - cls.HEADER_SIZE
        if datalen not in (MVPN_SHAREDJOIN_IPV4_LENGTH, MVPN_SHAREDJOIN_IPV6_LENGTH):  # IPv4 or IPv6
            raise Notify(3, 5, f'Invalid C-Multicast Route length ({datalen} bytes).')

        # Validate source IP length (offset +2 for header)
        cursor = 14  # 2 (header) + 8 (RD) + 4 (Source AS)
        sourceiplen = int(packed[cursor] / 8)
        cursor += 1
        if sourceiplen != IPv4.BYTES and sourceiplen != IPv6.BYTES:
            raise Notify(
                3,
                5,
                f'Invalid C-Multicast Route length ({sourceiplen * 8} bits). Expected 32 bits (IPv4) or 128 bits (IPv6).',
            )
        cursor += sourceiplen

        # Validate group IP length
        groupiplen = int(packed[cursor] / 8)
        if groupiplen != IPv4.BYTES and groupiplen != IPv6.BYTES:
            raise Notify(
                3,
                5,
                f'Invalid C-Multicast Route length ({groupiplen * 8} bits). Expected 32 bits (IPv4) or 128 bits (IPv6).',
            )

        return cls(packed, afi)

    def json(self, announced: bool = True, compact: bool | None = None) -> str:
        content = ' "code": %d, ' % self.CODE
        content += '"parsed": true, '
        content += '"raw": "{}", '.format(self._raw())
        content += '"name": "{}", '.format(self.NAME)
        content += '{}, '.format(self.rd.json())
        content += '"source-as": "{}", '.format(str(self.source_as))
        content += '"source": "{}", '.format(str(self.source))
        content += '"group": "{}"'.format(str(self.group))
        return '{{{}}}'.format(content)



"""mprnlri.py

Created by Thomas Mangin on 2009-11-05.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from __future__ import annotations

from struct import unpack
from typing import TYPE_CHECKING, Iterator

from exabgp.util.types import Buffer

if TYPE_CHECKING:
    from exabgp.bgp.message.open.capability.negotiated import Negotiated
    from exabgp.bgp.message.update.collection import RoutedNLRI

from exabgp.bgp.message.action import Action
from exabgp.bgp.message.notification import Notify
from exabgp.bgp.message.open.capability import Negotiated
from exabgp.bgp.message.update.attribute import Attribute, NextHop
from exabgp.bgp.message.update.nlri import NLRI
from exabgp.protocol.family import AFI, SAFI, Family

# ==================================================== MP Reachable NLRI (14)
#


@Attribute.register()
class MPRNLRI(Attribute, Family):
    """Wire-format MP_REACH_NLRI attribute container.

    Stores raw wire bytes and yields NLRIs lazily via __iter__.
    For semantic operations (building/packing), use MPNLRICollection.
    """

    FLAG = Attribute.Flag.OPTIONAL
    ID = Attribute.CODE.MP_REACH_NLRI
    NO_DUPLICATE = True

    def __init__(self, packed: Buffer, addpath: bool) -> None:
        """Create MPRNLRI from wire-format bytes.

        Args:
            packed: Wire-format payload (after attribute header)
            addpath: Whether AddPath is enabled for this AFI/SAFI
        """
        self._packed = packed
        self._addpath = addpath
        # Initialize Family from packed data
        _afi = unpack('!H', packed[:2])[0]
        _safi = packed[2]
        Family.__init__(self, AFI.from_int(_afi), SAFI.from_int(_safi))

    @property
    def packed(self) -> bytes:
        """Raw wire-format bytes."""
        return bytes(self._packed)

    def _parse_nexthop_and_nlris(self) -> tuple[Buffer | None, Iterator[NLRI]]:
        """Parse wire format, returning (nexthop_bytes, nlri_iterator).

        Internal method that separates nexthop parsing from NLRI parsing.
        """
        data = self._packed

        # -- Reading AFI/SAFI (already done in __init__ for Family)
        offset = 3

        # -- Reading length of next-hop
        len_nh = data[offset]
        offset += 1

        if (self.afi, self.safi) not in Family.size:
            raise Notify(3, 0, 'unsupported {} {}'.format(self.afi, self.safi))

        length, rd = Family.size[(self.afi, self.safi)]

        size = len_nh - rd

        # Parse nexthops
        nhs = data[offset + rd : offset + rd + size]
        nexthops = [nhs[pos : pos + 16] for pos in range(0, len(nhs), 16)]
        nexthop_bytes = nexthops[0] if nexthops else None

        offset += len_nh

        # Skip reserved byte
        offset += 1

        # Reading the NLRIs
        nlri_data = data[offset:]

        def nlri_generator() -> Iterator[NLRI]:
            nonlocal nlri_data
            while nlri_data:
                nlri_result, left_result = NLRI.unpack_nlri(
                    self.afi, self.safi, nlri_data, Action.ANNOUNCE, self._addpath, Negotiated.UNSET
                )

                if nlri_result is not NLRI.INVALID:
                    yield nlri_result

                if left_result == nlri_data:
                    raise RuntimeError('sub-calls should consume data')

                nlri_data = left_result

        return nexthop_bytes, nlri_generator()

    def __iter__(self) -> Iterator[NLRI]:
        """Yield NLRIs from wire format.

        Generator that yields NLRIs one by one, parsing lazily.
        Note: Does NOT set nlri.nexthop - use iter_routed() for RoutedNLRI.
        """
        _, nlri_iter = self._parse_nexthop_and_nlris()
        yield from nlri_iter

    def iter_routed(self) -> Iterator['RoutedNLRI']:
        """Yield RoutedNLRI from wire format.

        Generator that yields RoutedNLRI (nlri + nexthop) one by one.
        This is the preferred method for getting announces with nexthop.
        """
        from exabgp.bgp.message.update.collection import RoutedNLRI
        from exabgp.protocol.ip import IP

        nexthop_bytes, nlri_iter = self._parse_nexthop_and_nlris()
        # Convert NextHop (Attribute) to IP for RoutedNLRI
        nexthop: IP
        if nexthop_bytes is None:
            nexthop = IP.NoNextHop
        else:
            nexthop_attr = NextHop.unpack_attribute(nexthop_bytes, Negotiated.UNSET)
            if nexthop_attr is NextHop.UNSET:
                nexthop = IP.NoNextHop
            elif isinstance(nexthop_attr, NextHop):
                nexthop = IP.create_ip(nexthop_attr.pack_ip())
            else:
                nexthop = IP.NoNextHop
        for nlri in nlri_iter:
            yield RoutedNLRI(nlri, nexthop)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, MPRNLRI):
            return False
        return self.ID == other.ID and self.FLAG == other.FLAG and self._packed == other._packed

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __len__(self) -> int:
        # Return 0 to indicate unknown length - use list(mprnlri) to iterate
        return 0

    def __repr__(self) -> str:
        return 'MP_REACH_NLRI for %s %s' % (self.afi, self.safi)

    @classmethod
    def unpack_attribute(cls, data: Buffer, negotiated: Negotiated) -> Attribute:
        """Unpack MPRNLRI from wire format.

        Validates the data and creates an MPRNLRI instance storing the wire bytes.
        NLRIs are parsed lazily when iterating over the instance.
        """
        # MP_REACH_NLRI minimum: AFI(2) + SAFI(1) + NH_len(1) + reserved(1) = 5 bytes
        if len(data) < 5:
            raise Notify(3, 9, f'MP_REACH_NLRI too short: need at least 5 bytes, got {len(data)}')

        # -- Reading AFI/SAFI for validation
        _afi, _safi = unpack('!HB', data[:3])
        afi, safi = AFI.from_int(_afi), SAFI.from_int(_safi)
        offset = 3

        # we do not want to accept unknown families
        if negotiated and (afi, safi) not in negotiated.families:
            raise Notify(3, 0, 'presented a non-negotiated family {}/{}'.format(afi, safi))

        # -- Reading length of next-hop
        len_nh = data[offset]
        offset += 1

        # Validate we have enough data for next-hop + reserved byte
        if len(data) < offset + len_nh + 1:
            raise Notify(3, 9, f'MP_REACH_NLRI truncated: need {offset + len_nh + 1} bytes, got {len(data)}')

        if (afi, safi) not in Family.size:
            raise Notify(3, 0, 'unsupported {} {}'.format(afi, safi))

        length, rd = Family.size[(afi, safi)]

        # Link-Local Next Hop Capability (code 77) validation:
        # - 16-byte IPv6 nexthops are valid (could be global or link-local)
        # - With LLNH negotiated, 16-byte link-local (fe80::/10) is explicitly allowed
        # - Semantic interpretation of 16-byte NH depends on LLNH negotiation
        if negotiated.nexthop:
            if len_nh in (16, 32, 24):
                nh_afi = AFI.ipv6
            elif len_nh in (4, 12):
                nh_afi = AFI.ipv4
            else:
                raise Notify(
                    3, 0, 'unsupported family {} {} with extended next-hop capability enabled'.format(afi, safi)
                )
            length, _ = Family.size[(nh_afi, safi)]

        if len_nh not in length:
            raise Notify(
                3,
                0,
                'invalid %s %s next-hop length %d expected %s'
                % (afi, safi, len_nh, ' or '.join(str(_) for _ in length)),
            )

        # check the RD is well zero
        if rd and sum([int(_) for _ in data[offset : offset + 8]]) != 0:
            raise Notify(3, 0, "MP_REACH_NLRI next-hop's route-distinguisher must be zero")

        offset += len_nh

        # Skip a reserved bit as someone had to bug us !
        reserved = data[offset]
        offset += 1

        if reserved != 0:
            raise Notify(3, 0, 'the reserved bit of MP_REACH_NLRI is not zero')

        # Verify there's NLRI data
        if offset >= len(data):
            raise Notify(3, 0, 'No data to decode in an MPREACHNLRI but it is not an EOR %d/%d' % (afi, safi))

        # Get addpath flag for lazy NLRI parsing
        addpath = negotiated.required(afi, safi)

        # Store wire bytes and addpath flag - NLRIs parsed lazily
        return cls(data, addpath)


# Create empty MPRNLRI with minimal packed structure
# AFI(2) + SAFI(1) + NH_LEN(1)=0 + RESERVED(1)=0 = 5 bytes minimum
_EMPTY_PACKED = AFI.undefined.pack_afi() + SAFI.undefined.pack_safi() + bytes([0, 0])
EMPTY_MPRNLRI = MPRNLRI(_EMPTY_PACKED, addpath=False)



"""rtc.py

Created by Thomas Morin on 2014-06-10.
Copyright (c) 2014-2017 Orange. All rights reserved.
Copyright (c) 2014-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from __future__ import annotations

from struct import pack, unpack
from typing import TYPE_CHECKING, Any, Type, TypeVar

from exabgp.util.types import Buffer

if TYPE_CHECKING:
    from exabgp.bgp.message.open.capability.negotiated import Negotiated

from exabgp.bgp.message import Action
from exabgp.bgp.message.open.asn import ASN
from exabgp.bgp.message.update.attribute import Attribute
from exabgp.bgp.message.update.attribute.community.extended import RouteTarget
from exabgp.bgp.message.update.nlri.nlri import NLRI
from exabgp.protocol.family import AFI, SAFI, Family

T = TypeVar('T', bound='RTC')


@NLRI.register(AFI.ipv4, SAFI.rtc)
class RTC(NLRI):
    """RTC (Route Target Constraint) NLRI using packed-bytes-first pattern.

    Wire format (13 bytes for full RTC, 1 byte for wildcard):
    [length(1)] [origin(4)] [rt(8)]
     0:1         1:5         5:13

    - length: Length in bits (96 for full RTC, 0 for wildcard)
    - origin: Origin ASN (4 bytes, big-endian)
    - rt: RouteTarget with flags reset (8 bytes)

    Limitation: RFC 4684 prefix-based RTC filtering (variable length with partial RT)
    is not yet implemented - only full RTC constraints are supported.
    """

    __slots__ = ()  # Only _packed needed, inherited from NLRI

    # Wire format constants
    PACKED_LENGTH_FULL = 13  # 1 + 4 + 8
    PACKED_LENGTH_WILDCARD = 1

    # Fixed AFI/SAFI for this single-family NLRI type
    @property
    def afi(self) -> AFI:
        return AFI.ipv4

    @property
    def safi(self) -> SAFI:
        return SAFI.rtc

    def __init__(self, packed: Buffer) -> None:
        """Create an RTC NLRI from packed wire-format bytes.

        Args:
            packed: Wire format bytes (13 bytes for full RTC, 1 byte for wildcard)
        """
        NLRI.__init__(self, AFI.ipv4, SAFI.rtc)
        self._packed: Buffer = packed

    @property
    def origin(self) -> ASN:
        """Origin ASN - unpacked from wire bytes on access."""
        if len(self._packed) < 5:
            return ASN(0)
        return ASN(unpack('!L', self._packed[1:5])[0])

    @property
    def rt(self) -> RouteTarget | None:
        """RouteTarget - lazily unpacked from wire bytes on access."""
        from typing import cast

        if len(self._packed) < 13:
            return None
        # RT is stored with flags already reset, use unpack_attribute for proper subclass dispatch
        return cast(RouteTarget, RouteTarget.unpack_attribute(self._packed[5:13], None))

    @classmethod
    def make_rtc(
        cls,
        origin: ASN,
        rt: RouteTarget | None,
    ) -> 'RTC':
        """Factory method to create an RTC NLRI from components.

        Args:
            origin: Origin ASN
            rt: RouteTarget or None for wildcard

        Returns:
            New RTC instance

        Note: nexthop is stored in Route, not NLRI. Pass nexthop to Route constructor.
        """
        if rt is not None:
            packed_rt = rt._packed
            # Length in bits: (4 bytes origin + 8 bytes RT) * 8 = 96
            packed = pack('!BL', 96, int(origin)) + bytes([RTC.resetFlags(packed_rt[0])]) + packed_rt[1:]
        else:
            packed = pack('!B', 0)

        instance = cls(packed)
        return instance

    def feedback(self, action: Action) -> str:
        # Nexthop validation handled by Route.feedback()
        return ''

    def __len__(self) -> int:
        # Length in bits: for wildcard (single 0 byte), return 1
        # For full RTC: length is stored at byte 0 (96 bits = (4+8)*8)
        return self._packed[0] if self._packed[0] != 0 else 1

    def __str__(self) -> str:
        if len(self._packed) >= 13:
            return 'rtc {}:{}'.format(self.origin, self.rt)
        return 'rtc wildcard'

    def __repr__(self) -> str:
        return str(self)

    def __copy__(self) -> 'RTC':
        new = self.__class__.__new__(self.__class__)
        self._copy_nlri_slots(new)
        new._packed = self._packed  # bytes - immutable
        return new

    def __deepcopy__(self, memo: dict[Any, Any]) -> 'RTC':
        new = self.__class__.__new__(self.__class__)
        memo[id(self)] = new
        self._deepcopy_nlri_slots(new, memo)
        new._packed = self._packed  # bytes - immutable
        return new

    @staticmethod
    def resetFlags(char: int) -> int:
        return char & ~(Attribute.Flag.TRANSITIVE | Attribute.Flag.OPTIONAL)

    def pack_nlri(self, negotiated: Negotiated) -> Buffer:
        """Pack NLRI - returns stored wire bytes directly (zero-copy)."""
        return self._packed

    def index(self) -> bytes:
        """Return unique index for this RTC NLRI."""
        return Family.index(self) + self._packed

    @classmethod
    def unpack_nlri(
        cls: Type[T], afi: AFI, safi: SAFI, bgp: Buffer, action: Action, addpath: Any, negotiated: Negotiated
    ) -> tuple[T, Buffer]:
        data = memoryview(bgp) if not isinstance(bgp, memoryview) else bgp
        # Note: afi/safi parameters are ignored - RTC is always ipv4/rtc
        length = data[0]

        if length == 0:
            nlri = cls(bytes(data[0:1]))
            return nlri, data[1:]

        if length < 8 * 4:
            raise Exception('incorrect RT length: %d (should be >=32,<=96)' % length)

        # Store complete wire format with flags reset on RT
        # Wire format: [length(1)][origin(4)][rt(8)]
        packed = (
            bytes(data[0:5])  # length + origin
            + bytes([RTC.resetFlags(data[5])])  # RT first byte with flags reset
            + bytes(data[6:13])  # RT remaining bytes
        )

        nlri = cls(packed)
        return nlri, data[13:]



"""prefix.py

Created by Diego Garcia del Rio on 2015-03-12.
Copyright (c) 2015 Alcatel-Lucent. All rights reserved.

Based on work by Thomas Morin on mac.py
Copyright (c) 2014-2017 Orange. All rights reserved.
Copyright (c) 2014-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from __future__ import annotations

from typing import ClassVar

from exabgp.bgp.message import Action
from exabgp.bgp.message.notification import Notify
from exabgp.bgp.message.update.nlri import NLRI
from exabgp.bgp.message.update.nlri.evpn.nlri import EVPN
from exabgp.bgp.message.update.nlri.qualifier import ESI, EthernetTag, Labels, RouteDistinguisher
from exabgp.bgp.message.update.nlri.qualifier.path import PathInfo
from exabgp.protocol.ip import IP
from exabgp.util.types import Buffer

# ------------ EVPN Prefix Advertisement NLRI ------------
# As described here:
# https://tools.ietf.org/html/draft-ietf-bess-evpn-prefix-advertisement-01

# +---------------------------------------+
# |      RD   (8 octets)                  |
# +---------------------------------------+
# |Ethernet Segment Identifier (10 octets)|
# +---------------------------------------+
# |  Ethernet Tag ID (4 octets)           |
# +---------------------------------------+
# |  IP Prefix Length (1 octet)           |
# +---------------------------------------+
# |  IP Prefix (4 or 16 octets)           |
# +---------------------------------------+
# |  GW IP Address (4 or 16 octets)       |
# +---------------------------------------+
# |  MPLS Label (3 octets)                |
# +---------------------------------------+
# total NLRI length is 34 bytes for IPv4 or 58 bytes for IPv6

# ======================================================================= Prefix

# https://tools.ietf.org/html/draft-rabadan-l2vpn-evpn-prefix-advertisement-03


@EVPN.register_evpn_route(code=5)
class Prefix(EVPN):
    """EVPN Route Type 5: IP Prefix Advertisement.

    Wire format: type(1) + length(1) + RD(8) + ESI(10) + ETag(4) + IPlen(1)
                 + IP(4/16) + GwIP(4/16) + Label(3)
    Total: 36 bytes for IPv4, 60 bytes for IPv6 (including 2-byte header)
    Uses packed-bytes-first pattern for zero-copy routing.
    """

    NAME: ClassVar[str] = 'IP Prefix Advertisement'
    SHORT_NAME: ClassVar[str] = 'PrfxAdv'

    def __init__(self, packed: Buffer) -> None:
        """Create Prefix from complete wire-format bytes.

        Args:
            packed: Complete wire format (type + length + payload)
        """
        EVPN.__init__(self, packed)

    @classmethod
    def make_prefix(
        cls,
        rd: RouteDistinguisher,
        esi: ESI,
        etag: EthernetTag,
        label: Labels | None,
        ip: IP,
        iplen: int,
        gwip: IP,
        action: Action = Action.UNSET,
        addpath: PathInfo = PathInfo.DISABLED,
    ) -> 'Prefix':
        """Factory method to create Prefix from semantic parameters.

        Packs fields into wire format immediately (packed-bytes-first pattern).
        Note: nexthop is not part of NLRI - set separately after creation.

        rd: a RouteDistinguisher
        esi: an EthernetSegmentIdentifier
        etag: an EthernetTag
        label: a LabelStackEntry
        ip: an IP address (dotted quad string notation)
        iplen: prefixlength for ip (defaults to 32)
        gwip: an IP address (dotted quad string notation)
        """
        label_to_use = label if label else Labels.NOLABEL
        payload = (
            bytes(rd.pack_rd())
            + esi.pack_esi()
            + etag.pack_etag()
            + bytes([iplen])
            + ip.pack_ip()
            + gwip.pack_ip()
            + label_to_use.pack_labels()
        )
        # Include type + length header for zero-copy pack
        packed = bytes([cls.CODE, len(payload)]) + payload
        instance = cls(packed)
        instance.addpath = addpath
        return instance

    # Wire format offsets (after 2-byte type+length header):
    # RD: 2-10, ESI: 10-20, ETag: 20-24, IPlen: 24, IP: 25+, GwIP: after IP, Label: after GwIP
    # Total: 36 bytes for IPv4, 60 bytes for IPv6

    @property
    def rd(self) -> RouteDistinguisher:
        """Route Distinguisher - unpacked from wire bytes."""
        return RouteDistinguisher.unpack_routedistinguisher(self._packed[2:10])

    @property
    def esi(self) -> ESI:
        """Ethernet Segment Identifier - unpacked from wire bytes."""
        return ESI.unpack_esi(self._packed[10:20])

    @property
    def etag(self) -> EthernetTag:
        """Ethernet Tag - unpacked from wire bytes."""
        return EthernetTag.unpack_etag(self._packed[20:24])

    @property
    def iplen(self) -> int:
        """IP prefix length - unpacked from wire bytes."""
        return self._packed[24]

    @property
    def ip(self) -> IP:
        """IP prefix - unpacked from wire bytes."""
        # IP address is either 4 or 16 bytes based on total length
        datalen = len(self._packed)
        if datalen == 36:  # IPv4: 2+8+10+4+1+4+4+3
            return IP.create_ip(self._packed[25:29])
        else:  # IPv6: 2+8+10+4+1+16+16+3 = 60
            return IP.create_ip(self._packed[25:41])

    @property
    def gwip(self) -> IP:
        """Gateway IP - unpacked from wire bytes."""
        datalen = len(self._packed)
        if datalen == 36:  # IPv4
            return IP.create_ip(self._packed[29:33])
        else:  # IPv6
            return IP.create_ip(self._packed[41:57])

    @property
    def label(self) -> Labels:
        """MPLS Labels - unpacked from wire bytes."""
        datalen = len(self._packed)
        if datalen == 36:  # IPv4
            return Labels.unpack_labels(self._packed[33:36])
        else:  # IPv6
            return Labels.unpack_labels(self._packed[57:60])

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Prefix):
            return False
        return (
            NLRI.__eq__(self, other)
            and self.CODE == other.CODE
            and self.rd == other.rd
            and self.etag == other.etag
            and self.ip == other.ip
            and self.iplen == other.iplen
        )
        # esi, label and gwip must not be compared

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __str__(self) -> str:
        return '{}:{}:{}:{}:{}{}:{}:{}'.format(
            self._prefix(),
            self.rd._str(),
            self.esi,
            self.etag,
            self.ip,
            '/%d' % self.iplen,
            self.gwip,
            self.label,
        )

    def __hash__(self) -> int:
        # esi, and label, gwip must *not* be part of the hash
        return hash('{}:{}:{}:{}'.format(self.rd, self.etag, self.ip, self.iplen))

    @classmethod
    def unpack_evpn(cls, packed: Buffer) -> EVPN:
        """Unpack Prefix from complete wire format bytes.

        Args:
            packed: Complete wire format (type + length + payload)

        Returns:
            Prefix instance with stored wire bytes
        """
        # Get the data length to understand if addresses are IPv4 or IPv6
        # Lengths include 2-byte header: 36 for IPv4 (34+2), 60 for IPv6 (58+2)
        datalen = len(packed)

        if datalen not in (36, 60):  # 36 for IPv4, 60 for IPv6
            raise Notify(
                3,
                5,
                'Data field length is given as %d, but EVPN route currently support only IPv4 or IPv6 (36 or 60)'
                % datalen,
            )

        return cls(packed)

    def json(self, announced: bool = True, compact: bool = False) -> str:
        content = ' "code": %d, ' % self.CODE
        content += '"parsed": true, '
        content += '"raw": "{}", '.format(self._raw())
        content += '"name": "{}", '.format(self.NAME)
        content += '{}, '.format(self.rd.json())
        content += '{}, '.format(self.esi.json())
        content += '{}, '.format(self.etag.json())
        content += '{}, '.format(self.label.json())
        content += '"ip": "{}", '.format(str(self.ip))
        content += '"iplen": %d, ' % self.iplen
        content += '"gateway": "{}" '.format(str(self.gwip))
        return '{{{}}}'.format(content)



"""sr/srgb.py

Created by Evelio Vila 2017-02-16
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
"""

from __future__ import annotations

import json
from struct import pack
from struct import unpack
from typing import ClassVar

from exabgp.bgp.message.update.attribute.sr.prefixsid import PrefixSid
from exabgp.util.types import Buffer

# 0                   1                   2                   3
# 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |     Type      |          Length               |    Flags      |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |     Flags     |
# +-+-+-+-+-+-+-+-+
#
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |         SRGB 1 (6 octets)                                     |
# |                               +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |                               |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |         SRGB n (6 octets)                                     |
# |                               +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |                               |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# 3.3.  Originator SRGB TLV


@PrefixSid.register_sr()
class SrGb:
    TLV: ClassVar[int] = 3
    # Length is the total length of the value portion of the TLV: 2 +
    # multiple of 6.
    LENGTH: ClassVar[int] = -1

    def __init__(self, packed: Buffer) -> None:
        # Payload format: Flags(2) + N * (Base(3) + Range(3))
        # Minimum: 2 bytes (flags only), remainder must be divisible by 6
        if len(packed) < 2 or (len(packed) - 2) % 6 != 0:
            raise ValueError(f'Invalid SRGB payload size: {len(packed)} bytes (must be 2 + N*6)')
        self._packed: Buffer = packed

    @classmethod
    def make_srgb(cls, srgbs: list[tuple[int, int]]) -> 'SrGb':
        """Factory method for semantic construction."""
        payload: bytes = pack('!H', 0)  # flags
        for b, r in srgbs:
            payload = payload + pack('!L', b)[1:] + pack('!L', r)[1:]
        return cls(payload)

    @property
    def srgbs(self) -> list[tuple[int, int]]:
        """List of (base, range) tuples unpacked from _packed."""
        result: list[tuple[int, int]] = []
        data = self._packed[2:]  # Skip flags
        while data:
            base: int = unpack('!L', bytes([0]) + data[:3])[0]
            srange: int = unpack('!L', bytes([0]) + data[3:6])[0]
            result.append((base, srange))
            data = data[6:]
        return result

    def __repr__(self) -> str:
        items: list[str] = []
        for base, srange in self.srgbs:
            items.append(f'( {base},{srange} )')
        joined: str = ', '.join(items)
        return f'[ {joined} ]'

    def pack_tlv(self) -> bytes:
        return pack('!B', self.TLV) + pack('!H', len(self._packed)) + self._packed

    @classmethod
    def unpack_attribute(cls, data: Buffer, length: int) -> SrGb:
        # Validation happens in __init__
        return cls(data)

    def json(self, compact: bool | None = None) -> str:
        return f'"sr-srgbs": {json.dumps(self.srgbs)}'



"""admingroup.py

Created by Evelio Vila on 2016-12-01.
Copyright (c) 2014-2017 Exa Networks. All rights reserved.
"""

from __future__ import annotations

from struct import pack, unpack

from exabgp.bgp.message.update.attribute.bgpls.linkstate import LinkState
from exabgp.bgp.message.update.attribute.bgpls.linkstate import BaseLS
from exabgp.util.types import Buffer


@LinkState.register_lsid(tlv=1088, json_key='admin-group-mask', repr_name='Admin Group mask')
class AdminGroup(BaseLS):
    LEN = 4

    @property
    def content(self) -> int:
        """Unpack and return the admin group mask from packed bytes."""
        value: int = unpack('!L', self._packed[:4])[0]
        return value

    @classmethod
    def make_admingroup(cls, mask: int) -> AdminGroup:
        """Factory method to create AdminGroup from mask value."""
        return cls(pack('!I', mask))

    @classmethod
    def unpack_bgpls(cls, data: Buffer) -> AdminGroup:
        cls.check(data)
        return cls(data)



"""linkid.py

Created by Evelio Vila on 2016-11-26. eveliovila@gmail.com
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from __future__ import annotations

from struct import pack, unpack

from exabgp.util.types import Buffer


#       0                   1                   2                   3
#       0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
#      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#      |                  Link Local Identifier                        |
#      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#      |                  Link Remote Identifier                       |
#      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#      https://tools.ietf.org/html/rfc5307 sec 1.1
# ================================================================== Link Local/Remote Identifiers


class LinkIdentifier:
    def __init__(self, local_id: int, remote_id: int, packed: Buffer | None = None) -> None:
        self.local_id = local_id
        self.remote_id = remote_id
        self._packed = packed if packed is not None else pack('!LL', local_id, remote_id)

    @classmethod
    def unpack_linkid(cls, data: Buffer) -> 'LinkIdentifier':
        local_id = unpack('!L', data[:4])[0]
        remote_id = unpack('!L', data[4:8])[0]
        return cls(local_id=local_id, remote_id=remote_id, packed=data[:8])

    def json(self) -> str:
        return f'{{ "link-local-id": {self.local_id}, "link-remote-id": {self.remote_id} }}'

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, LinkIdentifier):
            return NotImplemented
        return self.local_id == other.local_id and self.remote_id == other.remote_id

    def __lt__(self, other: LinkIdentifier) -> bool:
        raise RuntimeError('Not implemented')

    def __le__(self, other: LinkIdentifier) -> bool:
        raise RuntimeError('Not implemented')

    def __gt__(self, other: LinkIdentifier) -> bool:
        raise RuntimeError('Not implemented')

    def __ge__(self, other: LinkIdentifier) -> bool:
        raise RuntimeError('Not implemented')

    def __str__(self) -> str:
        return ':'.join('{:02X}'.format(_) for _ in self._packed)

    def __repr__(self) -> str:
        return self.__str__()

    def __len__(self) -> int:
        if not self._packed:
            return 0
        return len(self._packed)

    def __hash__(self) -> int:
        return hash(str(self))

    def pack_tlv(self) -> Buffer:
        return self._packed



"""watchdog.py

Sentinel class for watchdog attribute.

Created by Thomas Mangin on 2009-11-05.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from __future__ import annotations


class Watchdog:
    """Watchdog attribute with name.

    Used as internal pseudo-attribute to track route announcements
    that should be controlled via watchdog commands.
    """

    __slots__ = ('_name',)

    def __init__(self, name: str) -> None:
        self._name = name

    @classmethod
    def _create_sentinel(cls) -> Watchdog:
        """Create the NoWatchdog sentinel instance."""
        instance = object.__new__(cls)
        instance._name = ''
        return instance

    @property
    def name(self) -> str:
        return self._name

    def __bool__(self) -> bool:
        return bool(self._name)

    def __str__(self) -> str:
        return self._name

    def __repr__(self) -> str:
        if self._name:
            return f'Watchdog({self._name!r})'
        return 'NoWatchdog'

    def __hash__(self) -> int:
        return hash(self._name)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Watchdog):
            return self._name == other._name
        return False


# Singleton sentinel - empty name is falsy
NoWatchdog: Watchdog = Watchdog._create_sentinel()



"""isd.py

Created by Takeru Hayasaka on 2023-01-21.
Copyright (c) 2023 BBSakura Networks Inc. All rights reserved.
"""

from __future__ import annotations

from struct import pack
from typing import TYPE_CHECKING, Any, ClassVar

from exabgp.util.types import Buffer

if TYPE_CHECKING:
    from exabgp.bgp.message.open.capability.negotiated import Negotiated

from exabgp.bgp.message import Action
from exabgp.bgp.message.update.nlri.mup.nlri import MUP
from exabgp.bgp.message.update.nlri.nlri import NLRI
from exabgp.bgp.message.update.nlri.qualifier import RouteDistinguisher
from exabgp.protocol.family import AFI, SAFI
from exabgp.protocol.ip import IP

# +-----------------------------------+
# |           RD  (8 octets)          |
# +-----------------------------------+
# |       Prefix Length (1 octet)     |
# +-----------------------------------+
# |        Prefix (variable)          |
# +-----------------------------------+


@MUP.register_mup_route(archtype=1, code=1)
class InterworkSegmentDiscoveryRoute(MUP):
    NAME: ClassVar[str] = 'InterworkSegmentDiscoveryRoute'
    SHORT_NAME: ClassVar[str] = 'ISD'

    # Wire format offsets (after 4-byte header: arch(1) + code(2) + length(1))
    HEADER_SIZE: ClassVar[int] = 4
    RD_OFFSET: ClassVar[int] = 4  # Bytes 4-11: RD (8 bytes)
    PREFIX_LEN_OFFSET: ClassVar[int] = 12  # Byte 12: prefix length
    PREFIX_OFFSET: ClassVar[int] = 13  # Bytes 13+: prefix

    def __init__(self, packed: Buffer, afi: AFI) -> None:
        """Create ISD with complete wire format.

        Args:
            packed: Complete wire format including 4-byte header
        """
        MUP.__init__(self, afi)
        self._packed: Buffer = packed

    @classmethod
    def make_isd(
        cls,
        rd: RouteDistinguisher,
        prefix_ip_len: int,
        prefix_ip: IP,
        afi: AFI,
    ) -> 'InterworkSegmentDiscoveryRoute':
        """Factory method to create ISD from semantic parameters."""
        offset = prefix_ip_len // 8
        remainder = prefix_ip_len % 8
        if remainder != 0:
            offset += 1

        prefix_ip_packed = prefix_ip.pack_ip()
        payload = bytes(rd.pack_rd()) + pack('!B', prefix_ip_len) + prefix_ip_packed[0:offset]
        # Include 4-byte header: arch(1) + code(2) + length(1) + payload
        packed = pack('!BHB', cls.ARCHTYPE, cls.CODE, len(payload)) + payload
        return cls(packed, afi)

    @property
    def rd(self) -> RouteDistinguisher:
        # Offset by 4-byte header: RD at bytes 4-11
        return RouteDistinguisher.unpack_routedistinguisher(self._packed[4:12])

    @property
    def prefix_ip_len(self) -> int:
        # Offset by 4-byte header: prefix_len at byte 12
        return self._packed[12]

    @property
    def prefix_ip(self) -> IP:
        size = 4 if self.afi != AFI.ipv6 else 16
        # Offset by 4-byte header: prefix at bytes 13+
        ip = self._packed[13:]
        padding = size - len(ip)
        if padding > 0:
            ip = bytes(ip) + bytes(padding)
        return IP.create_ip(ip)

    def index(self) -> bytes:
        return MUP.index(self)

    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, InterworkSegmentDiscoveryRoute)
            and self.rd == other.rd
            and self.prefix_ip_len == other.prefix_ip_len
            and self.prefix_ip == other.prefix_ip
        )

    def __ne__(self, other: Any) -> bool:
        return not self.__eq__(other)

    def __str__(self) -> str:
        return '{}:{}:{}{}'.format(self._prefix(), self.rd._str(), self.prefix_ip, '/%d' % self.prefix_ip_len)

    def __hash__(self) -> int:
        # Direct _packed hash - include afi since MUP supports both IPv4 and IPv6
        return hash((self.afi, self._packed))

    @classmethod
    def unpack_nlri(
        cls, afi: AFI, safi: SAFI, data: Buffer, action: Action, addpath: Any, negotiated: Negotiated
    ) -> tuple[NLRI, Buffer]:
        # Parent provides complete wire format including 4-byte header
        instance = cls(data, afi)
        return instance, b''

    def json(self, announced: bool = True, compact: bool | None = None) -> str:
        content = '"name": "{}", '.format(self.NAME)
        content += '"arch": %d, ' % self.ARCHTYPE
        content += '"code": %d, ' % self.CODE
        content += '"prefix_ip_len": %d, ' % self.prefix_ip_len
        content += '"prefix_ip": "{}", '.format(str(self.prefix_ip))
        content += self.rd.json()
        content += ', "raw": "{}"'.format(self._raw())
        return '{{ {} }}'.format(content)



"""tokeniser.py

Created by Thomas Mangin on 2015-06-05.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from __future__ import annotations

from collections import deque
from typing import TYPE_CHECKING, Generator, Iterable, Iterator

if TYPE_CHECKING:
    from exabgp.configuration.core.error import Error
    from exabgp.configuration.core.scope import Scope

from exabgp.configuration.core.format import tokens
from exabgp.protocol.family import AFI

# Minimum line length for parameters extraction
MIN_LINE_LENGTH_FOR_PARAMS = 2  # Minimum line tokens needed to have parameters


class Tokeniser:
    def __init__(self) -> None:
        self.next: deque[str] = deque()
        self.tokens: list[str] = []
        self.generator: Iterator[str] = iter([])
        self.announce: bool = True
        self.afi = AFI.undefined
        self.fname: str = ''
        self.consumed: int = 0

    def replenish(self, content: list[str]) -> 'Tokeniser':
        self.next.clear()
        self.tokens = content
        self.generator = iter(content)
        self.consumed = 0
        return self

    def clear(self) -> None:
        self.replenish([])
        self.announce = True

    def peek(self) -> str:
        """Peek at next token without incrementing consumed counter."""
        if self.next:
            return self.next[0]

        try:
            peaked = next(self.generator)
            self.next.append(peaked)
            return peaked
        except StopIteration:
            return ''

    def _get(self) -> str:
        """Get next token, incrementing consumed counter."""
        if self.next:
            self.consumed += 1
            return self.next.popleft()

        try:
            tok = next(self.generator)
            self.consumed += 1
            return tok
        except StopIteration:
            return ''

    def __call__(self) -> str:
        return self._get()

    def consume(self, name: str) -> None:
        next_tok = self._get()
        if next_tok != name:
            raise ValueError(f"expected '{name}' but found '{next_tok}' instead")

    def consume_if_match(self, name: str) -> bool:
        next_tok = self.peek()
        if next_tok == name:
            self._get()
            return True
        return False

    def remaining_string(self) -> str:
        """Get remaining tokens as the original command substring.

        Uses tokeniser.consumed to know how many words were consumed
        during dispatch, then extracts the remaining portion of
        the original command string.

        Args:
            tokeniser: Tokeniser with consumed count
            original: Original command string

        Returns:
            Remaining portion of original command after consumed words
        """

        return ' '.join([_ for _ in self.next] + [_ for _ in self.generator])


class Parser:
    @staticmethod
    def _off() -> Iterator[list[str]]:
        return iter([])

    def __init__(self, scope: Scope, error: Error) -> None:
        self.scope: Scope = scope
        self.error: Error = error
        self.finished = False
        self.number = 0
        self.line: list[str] = []
        self.tokeniser: Tokeniser = Tokeniser()
        self.end = ''
        self.index_column = 0
        self.index_line = 0
        self.fname = ''
        self.type = 'unset'

        self._tokens: Iterator[list[str]] = Parser._off()
        self._next: list[str] = []
        self._data: Generator[list[str], None, None] | None = None

    def clear(self) -> None:
        self.finished = False
        self.number = 0
        self.line = []
        self.tokeniser.clear()
        self.end = ''
        self.index_column = 0
        self.index_line = 0
        self.fname = ''
        self.type = 'unset'
        if self._data:
            self._set(self._data)

    def params(self) -> str:
        if len(self.line) <= MIN_LINE_LENGTH_FOR_PARAMS:
            return ''
        if self.end in ('{', '}', ';'):
            joined = "' '".join(self.line[1:-1])
            return f"'{joined}'"
        joined = "' '".join(self.line[1:])
        return f"'{joined}'"

    def _tokenise(self, iterator: Iterable[str]) -> Generator[list[str], None, None]:
        for parsed in tokens(iterator):
            words = [word for y, x, word in parsed]
            self.line = words  # Store the word list, not a joined string
            # ignore # lines
            # set Location information
            yield words

    def _set(self, function: Generator[list[str], None, None]) -> bool | str:
        try:
            self._tokens = function
            self._next = next(self._tokens)
        except OSError as exc:
            error = str(exc)
            if error.count(']'):
                self.error.set(error.split(']')[1].strip())
            else:
                self.error.set(error)
            self._tokens = Parser._off()
            self._next = []
            return self.error.set('issue setting the configuration parser')
        except StopIteration:
            self._tokens = Parser._off()
            self._next = []
            return self.error.set('issue setting the configuration parser, no data')
        return True

    def set_file(self, data: str) -> bool | str:
        def _source(fname: str) -> Generator[list[str], None, None]:
            with open(fname, 'r') as fileobject:

                def formated() -> Generator[str, None, None]:
                    line = ''
                    for current in fileobject:
                        self.index_line += 1
                        current = current.rstrip()
                        if current.endswith('\\'):
                            line += current[:-1]  # strip trailing backslash before concatenating
                            continue
                        if line:
                            yield line + current
                            line = ''
                            continue
                        yield current
                    if line:
                        yield line + current

                for _ in self._tokenise(formated()):
                    yield _

        self.type = 'file'
        self.fname = data
        self.tokeniser.fname = data
        return self._set(_source(data))

    def set_text(self, data: str) -> bool | str:
        def _source(data: str) -> Generator[list[str], None, None]:
            for _ in self._tokenise(data.split('\n')):
                yield _

        self.type = 'text'
        return self._set(_source(data))

    def set_api(self, line: str) -> bool | str:
        return self._set(self._tokenise(iter([line])))

    def set_action(self, command: str) -> None:
        if command != 'announce':
            self.tokeniser.announce = False

    def __call__(self) -> list[str]:
        self.number += 1
        try:
            self.line, self._next = self._next, next(self._tokens)
            self.end = self.line[-1]
        except StopIteration:
            if not self.finished:
                self.finished = True
                self.line, self._next = self._next, []
                self.end = self.line[-1]
            else:
                self.line = []
                self.end = ''

        # should we raise a Location if called with no more data ?
        self.tokeniser.replenish(self.line[:-1])

        return self.line



"""prefix.py

Created by Thomas Mangin on 2022-04-27.
Copyright (c) 2022-2022 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

#   https://www.rfc-editor.org/rfc/rfc7752.html#section-3.2.1.5

# The Multi-Topology ID (MT-ID) TLV carries one or more IS-IS or OSPF
# Multi-Topology IDs for a link, node, or prefix.

# Semantics of the IS-IS MT-ID are defined in Section 7.2 of RFC 5120
# [RFC5120].  Semantics of the OSPF MT-ID are defined in Section 3.7 of
# RFC 4915 [RFC4915].  If the value in the MT-ID TLV is derived from
# OSPF, then the upper 9 bits MUST be set to 0.  Bits R are reserved
# and SHOULD be set to 0 when originated and ignored on receipt.

# The format of the MT-ID TLV is shown in the following figure.

# 	0                   1                   2                   3
# 	0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
# 	+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# 	|              Type             |          Length=2*n           |
# 	+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# 	|R R R R|  Multi-Topology ID 1  |             ....             //
# 	+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# 	//             ....             |R R R R|  Multi-Topology ID n  |
# 	+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

# 				Figure 12: Multi-Topology ID TLV Format

# where Type is 263, Length is 2*n, and n is the number of MT-IDs
# carried in the TLV.

# The MT-ID TLV MAY be present in a Link Descriptor, a Prefix
# Descriptor, or the BGP-LS attribute of a Node NLRI.  In a Link or
# Prefix Descriptor, only a single MT-ID TLV containing the MT-ID of
# the topology where the link or the prefix is reachable is allowed.
# In case one wants to advertise multiple topologies for a given Link
# Descriptor or Prefix Descriptor, multiple NLRIs need to be generated
# where each NLRI contains an unique MT-ID.  In the BGP-LS attribute of
# a Node NLRI, one MT-ID TLV containing the array of MT-IDs of all
# topologies where the node is reachable is allowed.
# ================================================================== InterfaceAddress

from __future__ import annotations

import struct

from exabgp.util.types import Buffer


class MTID:
    def __init__(self, topologies: int, packed: Buffer) -> None:
        self.topologies = topologies
        self._packed = packed

    @classmethod
    def unpack_mtid(cls, data: Buffer) -> 'MTID':
        # tids = []
        # for i in range(0, len(data), 2):
        #     payload = struct.unpack('!H', data[i:i+2])[0]
        #     tids.append(payload & 0x0FFF)
        tids = struct.unpack('!H', data[:2])[0]
        return cls(tids, data)

    def json(self, compact: bool = False) -> str:
        return str(self.topologies)
        # tids = ', '.join(_ for _ in self.topologies)
        # return f'[{tids}]'

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, MTID):
            return NotImplemented
        return self.topologies == other.topologies

    def __lt__(self, other: MTID) -> bool:
        raise RuntimeError('Not implemented')

    def __le__(self, other: MTID) -> bool:
        raise RuntimeError('Not implemented')

    def __gt__(self, other: MTID) -> bool:
        raise RuntimeError('Not implemented')

    def __ge__(self, other: MTID) -> bool:
        raise RuntimeError('Not implemented')

    def __str__(self) -> str:
        return ':'.join('{:02X}'.format(_) for _ in self.pack_tlv())

    def __repr__(self) -> str:
        return self.__str__()

    def __len__(self) -> int:
        return len(self._packed)

    def __hash__(self) -> int:
        return hash(str(self))

    def pack_tlv(self) -> Buffer:
        return self._packed



"""routerid.py

Created by Thomas Mangin on 2012-07-17.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from __future__ import annotations

from exabgp.util.types import Buffer
from typing import Type

from exabgp.protocol.ip import IPv4

# ===================================================================== RouterID
#


class RouterID(IPv4):
    def __init__(self, ip: str) -> None:
        IPv4.__init__(self, IPv4.pton(ip))

    @classmethod
    def unpack_routerid(cls: Type[RouterID], data: Buffer) -> RouterID:
        return cls(IPv4.ntop(data))



"""rt_record.py

Created by Thomas Mangin on <unset>
Copyright (c) 2009-2022 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from __future__ import annotations

from typing import ClassVar, Type, TypeVar, cast

from exabgp.bgp.message.open.capability.negotiated import Negotiated
from exabgp.bgp.message.update.attribute.community.extended import ExtendedCommunity
from exabgp.bgp.message.update.attribute.community.extended import rt

# draft-ietf-bess-service-chaining

T = TypeVar('T', bound='RTRecord')


class RTRecord(rt.RouteTarget):
    COMMUNITY_SUBTYPE: ClassVar[int] = 0x13
    DESCRIPTION: ClassVar[str] = 'rtrecord'

    @classmethod
    def from_rt(cls: Type[T], route_target: rt.RouteTarget) -> T:
        packed = route_target.pack_attribute(Negotiated.UNSET)
        return cast(T, cls.unpack_attribute(bytes(packed[0:1]) + bytes([cls.COMMUNITY_SUBTYPE]) + bytes(packed[2:])))


@ExtendedCommunity.register_subtype
class RTRecordASN2Number(RTRecord, rt.RouteTargetASN2Number):
    pass


@ExtendedCommunity.register_subtype
class RTRecordIPNumber(RTRecord, rt.RouteTargetIPNumber):
    pass


@ExtendedCommunity.register_subtype
class RTRecordASN4Number(RTRecord, rt.RouteTargetASN4Number):
    pass



"""direction.py

Created by Thomas Mangin on 2010-01-15.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

# =================================================================== Direction
#

from __future__ import annotations

from enum import Enum


class Direction(Enum):
    IN = 1
    OUT = 2
